# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Any, Dict, List, Optional, Union

import numpy as np
import torch
import torch.nn as nn
from modelscope.models.builder import MODELS

from adaseq.data.span_utils import TypedSpan
from adaseq.metainfo import Models, Tasks
from adaseq.models.base import Model
from adaseq.modules.dropouts import WordDropout
from adaseq.modules.embedders import Embedder
from adaseq.modules.encoders import Encoder
from adaseq.modules.util import get_tokens_mask


class SinusoidalPositionEmbedding(nn.Module):
    """Sin-Cos Embedding.
    ref: https://spaces.ac.cn/archives/8265
    """

    def __init__(self, output_dim: int, merge_mode: str = 'add', custom_position_ids: bool = False):
        super(SinusoidalPositionEmbedding, self).__init__()
        self.output_dim = output_dim
        self.merge_mode = merge_mode
        self.custom_position_ids = custom_position_ids

    def forward(self, inputs):  # noqa
        if self.custom_position_ids:
            seq_len = inputs.shape[1]
            inputs, position_ids = inputs
            position_ids = position_ids.type(torch.float)
        else:
            input_shape = inputs.shape
            seq_len = input_shape[1]
            position_ids = torch.arange(seq_len).type(torch.float)[None]
        indices = torch.arange(self.output_dim // 2).type(torch.float)
        indices = torch.pow(10000.0, -2 * indices / self.output_dim)
        embeddings = torch.einsum('bn,d->bnd', position_ids, indices)
        embeddings = torch.stack([torch.sin(embeddings), torch.cos(embeddings)], dim=-1)
        embeddings = torch.reshape(embeddings, (-1, seq_len, self.output_dim))
        embeddings = embeddings.to(inputs.device)
        if self.merge_mode == 'add':
            return inputs + embeddings
        elif self.merge_mode == 'mul':
            return inputs * (embeddings + 1.0)
        elif self.merge_mode == 'zero':
            return embeddings


@MODELS.register_module(Tasks.named_entity_recognition, module_name=Models.global_pointer_model)
class GlobalPointerModel(Model):
    """GlobalPointer model.
    ref: https://arxiv.org/abs/2208.03054
    ref: https://github.com/xhw205/Efficient-GlobalPointer-torch
    """

    def __init__(
        self,
        id_to_label: Dict[int, str],
        embedder: Union[Embedder, Dict[str, Any]],
        encoder: Optional[Union[Encoder, Dict[str, Any]]] = None,
        token_ffn_out_width: int = -1,
        dropout: float = 0.0,
        word_dropout: bool = False,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.id_to_label = {int(k): v for k, v in id_to_label.items()}
        num_labels = len(id_to_label)
        self.num_classes = num_labels + 1

        if isinstance(embedder, Embedder):
            self.embedder = embedder
        else:
            self.embedder = Embedder.from_config(cfg_dict_or_path=embedder)
        hidden_size = self.embedder.get_output_dim()

        if encoder is None:
            self.encoder = None
        else:
            if isinstance(encoder, Encoder):
                self.encoder = encoder
            else:
                self.encoder = Encoder.from_config(encoder)
            assert hidden_size == self.encoder.get_input_dim()
            hidden_size = self.encoder.get_output_dim()

        self.token_ffn_out_width = token_ffn_out_width
        self.token_inner_embed_ffn = nn.Linear(hidden_size, token_ffn_out_width * 2)
        self.type_score_ffn = nn.Linear(hidden_size, num_labels * 2)
        self.pos_embed = SinusoidalPositionEmbedding(token_ffn_out_width, 'zero')

        self.use_dropout = dropout > 0.0
        if self.use_dropout:
            if word_dropout:
                self.dropout = WordDropout(dropout)
            else:
                self.dropout = nn.Dropout(dropout)

    def forward(
        self,
        tokens: Dict[str, Any],
        span_labels: Optional[torch.LongTensor] = None,
        meta: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:  # noqa
        """TODO: docstring"""
        entity_score = self._forward(tokens)
        if self.training:
            onehot = nn.functional.one_hot(span_labels, self.num_classes)
            label_matrix = onehot.permute(0, 3, 1, 2)[:, 1:, ...]
            loss = self._calculate_loss(entity_score, label_matrix)
            outputs = {'entity_score': entity_score, 'loss': loss}
        else:
            predicts = self.decode(entity_score)
            outputs = {'entity_score': entity_score, 'predicts': predicts}
        return outputs

    def _forward(self, tokens: Dict[str, Any]) -> torch.Tensor:
        x = self.embedder(**tokens)
        mask = get_tokens_mask(tokens, x.size(1))

        if self.use_dropout:
            x = self.dropout(x)

        if self.encoder is not None:
            x = self.encoder(x, mask)

            if self.use_dropout:
                x = self.dropout(x)

        token_inner_embed = self.token_inner_embed_ffn(x)
        start_token, end_token = token_inner_embed[..., ::2], token_inner_embed[..., 1::2]
        pos = self.pos_embed(token_inner_embed)
        cos_pos = pos[..., 1::2].repeat_interleave(2, dim=-1)
        sin_pos = pos[..., ::2].repeat_interleave(2, dim=-1)

        def add_position_embedding(in_embed, cos_pos, sin_pos):
            additional_part = torch.stack([-in_embed[..., 1::2], in_embed[..., ::2]], 3)
            additional_part = torch.reshape(additional_part, start_token.shape)
            output_embed = in_embed * cos_pos + additional_part * sin_pos
            return output_embed

        start_token = add_position_embedding(start_token, cos_pos, sin_pos)
        end_token = add_position_embedding(end_token, cos_pos, sin_pos)
        span_score = (
            torch.einsum('bmd,bnd->bmn', start_token, end_token) / self.token_ffn_out_width**0.5
        )
        typing_score = torch.einsum('bnh->bhn', self.type_score_ffn(x)) / 2
        entity_score = (
            span_score[:, None] + typing_score[:, ::2, None] + typing_score[:, 1::2, :, None]
        )  # [:, None] 增加一个维度

        entity_score = self._add_mask_tril(entity_score, mask=mask)
        return entity_score

    def _calculate_loss(self, entity_score, targets) -> torch.Tensor:
        """
        targets : (batch_size, num_classes, seq_len, seq_len)
        entity_score : (batch_size, num_classes, seq_len, seq_len)
        """
        batch_size, num_classes = entity_score.shape[:2]
        targets = targets.reshape(batch_size * num_classes, -1)
        entity_score = entity_score.reshape(batch_size * num_classes, -1)
        loss = self.multilabel_categorical_crossentropy(targets, entity_score)
        return loss

    def _sequence_masking(self, x, mask, value='-inf', axis=None) -> torch.Tensor:
        """Mask X according to the mask."""

        if mask is None:
            return x
        else:
            if value == '-inf':
                value = -1e12
            elif value == 'inf':
                value = 1e12
            assert axis > 0, 'axis must be greater than 0'
            for _ in range(axis - 1):
                mask = torch.unsqueeze(mask, 1)
            for _ in range(x.ndim - mask.ndim):
                mask = torch.unsqueeze(mask, mask.ndim)
            # return x * mask + value * (1 - mask)
            return x * mask + value * ~mask

    def _add_mask_tril(self, entity_score, mask):
        entity_score = self._sequence_masking(entity_score, mask, '-inf', entity_score.ndim - 2)
        entity_score = self._sequence_masking(entity_score, mask, '-inf', entity_score.ndim - 1)
        # 排除下三角
        mask = torch.tril(torch.ones_like(entity_score), diagonal=-1)
        entity_score = entity_score - mask * 1e12
        return entity_score

    def multilabel_categorical_crossentropy(self, targets, entity_score):
        """Multi-label cross entropy loss.
        https://kexue.fm/archives/7359
        """
        entity_score = (1 - 2 * targets) * entity_score  # -1 -> pos classes, 1 -> neg classes
        entity_score_neg = entity_score - targets * 1e12  # mask the pred outputs of pos classes
        entity_score_pos = (
            entity_score - (1 - targets) * 1e12
        )  # mask the pred outputs of neg classes
        zeros = torch.zeros_like(entity_score[..., :1])
        entity_score_neg = torch.cat([entity_score_neg, zeros], dim=-1)
        entity_score_pos = torch.cat([entity_score_pos, zeros], dim=-1)
        neg_loss = torch.logsumexp(entity_score_neg, dim=-1)
        pos_loss = torch.logsumexp(entity_score_pos, dim=-1)

        return (neg_loss + pos_loss).mean()

    def decode(self, entity_scores: torch.Tensor) -> List[List[TypedSpan]]:  # noqa
        entity_scores = entity_scores.detach().cpu().numpy()
        batch = list()
        for score_matrix in entity_scores:
            entities = [
                (start, end + 1, self.id_to_label[type_id])
                for type_id, start, end in zip(*np.where(score_matrix > 0))  # type: ignore
            ]
            batch.append(entities)
        return batch



# Copyright (c) Alibaba, Inc. and its affiliates.

from typing import Any, Dict

from modelscope.trainers.lrscheduler.builder import (
    build_lr_scheduler as ms_build_lr_scheduler,
)
from modelscope.utils.config import Config
from transformers import optimization


def build_lr_scheduler(config: Config, total_steps: int, default_args: Dict[str, Any]):
    """
    Build lr scheduler, `constant` by default.
    """
    if config is None:
        config = dict(type='constant')

    name = config.get('type')
    warmup_rate = config.get('warmup_rate', 0.0)

    # transformers lr_scheduler
    if name in optimization.TYPE_TO_SCHEDULER_FUNCTION:
        return optimization.get_scheduler(
            name,
            default_args['optimizer'],
            num_warmup_steps=int(total_steps * warmup_rate),
            num_training_steps=total_steps,
        )

    # torch lr_scheduler
    else:
        return ms_build_lr_scheduler(config, default_args)



import torch
from torch import nn
from transformers.activations import ACT2FN


# import from huggingface transformers
class PredictionHeadTransform(nn.Module):
    """TODO"""

    def __init__(self, config):
        super().__init__()
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        if isinstance(config.hidden_act, str):
            self.transform_act_fn = ACT2FN[config.hidden_act]
        else:
            self.transform_act_fn = config.hidden_act
        self.LayerNorm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:  # noqa
        hidden_states = self.dense(hidden_states)
        hidden_states = self.transform_act_fn(hidden_states)
        hidden_states = self.LayerNorm(hidden_states)
        return hidden_states


class MLMPredictionHead(nn.Module):
    """TODO"""

    def __init__(self, config):
        super().__init__()
        self.transform = PredictionHeadTransform(config)

        # The output weights are the same as the input embeddings, but there is
        # an output-only bias for each token.
        self.decoder = nn.Linear(config.hidden_size, config.vocab_size, bias=False)

        self.bias = nn.Parameter(torch.zeros(config.vocab_size))

        # Need a link between the two variables so that the bias is correctly resized with `resize_token_embeddings`
        self.decoder.bias = self.bias

    def forward(self, hidden_states):  # noqa
        hidden_states = self.transform(hidden_states)
        hidden_states = self.decoder(hidden_states)
        return hidden_states


class OnlyMLMHead(nn.Module):
    """TODO"""

    def __init__(self, config):
        super().__init__()
        self.predictions = MLMPredictionHead(config)

    def forward(self, sequence_output: torch.Tensor) -> torch.Tensor:  # noqa
        prediction_scores = self.predictions(sequence_output)
        return prediction_scores



# Copyright (c) Alibaba, Inc. and its affiliates.

import torch


class Biaffine(torch.nn.Module):
    """Biaffine Attention"""

    def __init__(self, in_features: int, out_features: int, bias=(True, True)):
        super().__init__()
        self.in_features = in_features  # mlp_arc_size / mlp_label_size
        self.out_features = out_features  # 1 / rel_size
        self.bias = bias

        # arc: mlp_size
        # label: mlp_size + 1
        self.linear_input_size = in_features + bias[0]
        # arc: mlp_size * 1
        # label: (mlp_size + 1) * rel_size
        self.linear_output_size = out_features * (in_features + bias[1])

        self.linear = torch.nn.Linear(
            in_features=self.linear_input_size, out_features=self.linear_output_size, bias=False
        )
        # self.reset_params()

    # def reset_params(self):
    #     nn.init.xavier_uniform_(self.linear.weight)

    def forward(self, input1, input2):  # noqa: D102
        batch_size, len1, dim1 = input1.size()
        batch_size, len2, dim2 = input2.size()

        if self.bias[0]:
            ones = input1.data.new_ones(batch_size, len1, 1)
            input1 = torch.cat((input1, ones), dim=-1)
            # dim1 += 1
        if self.bias[1]:
            ones = input2.data.new_ones(batch_size, len2, 1)
            input2 = torch.cat((input2, ones), dim=-1)
            # dim2 += 1

        # (bz, len1, dim1+1) -> (bz, len1, linear_output_size)
        affine = self.linear(input1)

        # (bz, len1 * self.out_features, dim2)
        affine = affine.reshape(batch_size, len1 * self.out_features, -1)

        # (bz, len1 * out_features, dim2) * (bz, dim2, len2)
        # -> (bz, len1 * out_features, len2) -> (bz, len2, len1 * out_features)
        biaffine = torch.bmm(affine, input2.transpose(1, 2)).transpose(1, 2).contiguous()

        # (bz, len2, len1, out_features)    # out_features: 1 or rel_size
        biaffine = biaffine.reshape((batch_size, len2, len1, -1)).squeeze(-1)

        return biaffine



# Copyright (c) Alibaba, Inc. and its affiliates.
from .multilabel_typing_data_collator_with_padding import (
    MultiLabelSpanTypingDataCollatorWithPadding,
)
from .pretraining_data_collator_with_padding import PretrainingDataCollatorWithPadding
from .sequence_labeling_data_collator_with_padding import (
    SequenceLabelingDataCollatorWithPadding,
)
from .span_extraction_data_collator_with_padding import (
    SpanExtractionDataCollatorWithPadding,
)
from .twostage_data_collator_with_padding import TwostageDataCollatorWithPadding



import torch


def bio2span(sequences, device, id_to_label):
    """label sequence to span format"""

    bsz = sequences.size(0)

    all_spans = [[] for _ in range(bsz)]

    for b in range(len(all_spans)):
        all_spans[b] = convert2span(sequences[b].tolist(), id_to_label)

    max_num_spans = max(len(i) for i in all_spans)

    if max_num_spans == 0:
        max_num_spans = 1

    pred_mentions = [i + (max_num_spans - len(i)) * [[-1, -1]] for i in all_spans]

    pred_mentions = torch.tensor(pred_mentions).to(device)
    pred_mask = pred_mentions[..., 0] != -1

    return pred_mentions, pred_mask


def convert2span(label_list, id_to_label):
    """convert label sequence to span"""
    span_list = []
    i = 0
    label_list = [id_to_label[item] for item in label_list]
    label_list = fix_tag_sequence_error(label_list)
    while i < len(label_list):
        if label_list[i][0] == 'B' or label_list[i][0] == 'I':
            start_idx = i
            i += 1
            while i < len(label_list) and not (label_list[i][0] == 'O' or label_list[i][0] == 'E'):
                i += 1
            if i < len(label_list):
                end_idx = i if label_list[i][0] == 'E' else i - 1
                # Looks like a good trick
                span_list.append([start_idx, end_idx])
                i += 1
            else:
                span_list.append([start_idx, i - 1])

        elif label_list[i][0] == 'S' or label_list[i][0] == 'E':
            span_list.append([i, i])
            i += 1
        else:
            i += 1

    return span_list


def fix_tag_sequence_error(tag_seq):
    """fix label sequence errors"""
    fixed_tag_seq = []
    last_type = 'O'
    for t in tag_seq:
        if t.startswith('E-'):
            t = t.replace('E-', 'I-')
        if t.startswith('S-'):
            t = t.replace('S-', 'B-')
        if t == 'O':
            fixed_tag_seq.append(t)
        elif t.startswith('B-'):
            fixed_tag_seq.append('B-' + t[2:])
            last_type = t[2:]
        elif t.startswith('I-'):
            cur_type = t[2:]
            if cur_type == last_type:
                fixed_tag_seq.append('I-' + cur_type)
            else:
                fixed_tag_seq.append('B-' + cur_type)
                last_type = cur_type
    return fixed_tag_seq



# Copyright (c) Alibaba, Inc. and its affiliates.
import json
from typing import Dict

from modelscope.metrics.builder import METRICS
from modelscope.trainers.parallel.utils import is_parallel
from modelscope.utils.tensor_utils import torch_nested_detach, torch_nested_numpify

from adaseq.data.constant import PAD_LABEL_ID
from adaseq.metainfo import DatasetDumpers

from .base import DatasetDumper


@METRICS.register_module(module_name=DatasetDumpers.ner_dumper)
class NamedEntityRecognitionDatasetDumper(DatasetDumper):
    """Named Entity Recognition dumper."""

    def __init__(self, model_type: str, dump_format: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_type = model_type
        self.dump_format = dump_format

    def add(self, outputs: Dict, inputs: Dict):
        """Only support sequence_labeling models now."""
        if self.model_type == 'sequence_labeling':
            self._add_sequence_labeling_data(outputs, inputs)
        elif self.model_type == 'span_based':
            self._add_span_based_data(outputs, inputs)
        else:
            raise NotImplementedError

    def dump(self):
        """Only support dump CoNLL format and jsonline now."""
        if self.dump_format == 'conll':
            self._dump_to_conll()
        elif self.dump_format == 'jsonline':
            self._dump_to_jsonline()
        else:
            raise NotImplementedError

    def _add_sequence_labeling_data(self, outputs: Dict, inputs: Dict):
        if is_parallel(self.trainer.model):
            id2label = self.trainer.model.module.id_to_label
        else:
            id2label = self.trainer.model.id_to_label
        batch_meta = inputs['meta']
        batch_labels = torch_nested_numpify(torch_nested_detach(inputs['label_ids'])).tolist()
        batch_predicts = torch_nested_numpify(torch_nested_detach(outputs['predicts'])).tolist()
        for meta, labels, predicts in zip(batch_meta, batch_labels, batch_predicts):
            self.data.append(
                {
                    'tokens': meta['tokens'],
                    'labels': [id2label[x] for x in labels if x != PAD_LABEL_ID],
                    'predicts': [id2label[x] for x in predicts if x != PAD_LABEL_ID],
                }
            )

    def _add_span_based_data(self, outputs: Dict, inputs: Dict):
        for meta, predicts in zip(inputs['meta'], outputs['predicts']):
            obj = dict(tokens=meta['tokens'], spans=meta['spans'], predicts=predicts)
            self.data.append(obj)

    def _dump_to_conll(self):
        with open(self.save_path, 'w', encoding='utf8') as fout:
            for example in self.data:
                for i in range(len(example['labels'])):
                    print(
                        example['tokens'][i],
                        example['labels'][i],
                        example['predicts'][i],
                        sep='\t',
                        file=fout,
                    )
                print('', file=fout)

    def _dump_to_jsonline(self):
        with open(self.save_path, 'w', encoding='utf8') as file:
            for example in self.data:
                file.write(json.dumps(example, ensure_ascii=False) + '\n')



# Copyright (c) Alibaba, Inc. and its affiliates.
import logging
from typing import Any, Dict, List, Union

from modelscope.preprocessors.builder import PREPROCESSORS
from modelscope.utils.constant import Fields

from adaseq.data.constant import NON_ENTITY_LABEL, PARTIAL_LABEL, PARTIAL_LABEL_ID
from adaseq.metainfo import Preprocessors

from .nlp_preprocessor import NLPPreprocessor

logger = logging.getLogger(__name__)


@PREPROCESSORS.register_module(Fields.nlp, module_name=Preprocessors.sequence_labeling_preprocessor)
class SequenceLabelingPreprocessor(NLPPreprocessor):
    """Preprocessor for Sequence Labeling"""

    def __init__(
        self, model_dir: str, labels: List[str], tag_scheme: str = 'BIOES', **kwargs
    ) -> None:
        self.tag_scheme = tag_scheme.upper()
        if not self._is_valid_tag_scheme(self.tag_scheme):
            raise ValueError('Invalid tag scheme! Options: [BIO, BIOES]')
        # self.tag_scheme = self._determine_tag_scheme_from_labels(labels)
        label_to_id = self._gen_label_to_id_with_bio(labels, self.tag_scheme)
        logger.info('label_to_id: ' + str(label_to_id))
        super().__init__(model_dir, label_to_id=label_to_id, return_offsets=True, **kwargs)

    def __call__(self, data: Union[str, List, Dict]) -> Dict[str, Any]:
        """prepare inputs for Sequence Labeling models."""
        output = super().__call__(data)
        if isinstance(data, Dict) and 'spans' in data:
            length = len(output['tokens']['mask']) - 2 * int(self.add_special_tokens)
            labels = self._spans_to_bio_labels(data['spans'], length, self.tag_scheme)
            output['label_ids'] = [
                PARTIAL_LABEL_ID if labels[i] == PARTIAL_LABEL else self.label_to_id[labels[i]]
                for i in range(length)
            ]
        return output

    @staticmethod
    def _is_valid_tag_scheme(tag_scheme: str):
        return tag_scheme in ['BIO', 'BIOES']

    @staticmethod
    def _determine_tag_scheme_from_labels(labels: List[str]) -> str:
        tag_scheme = 'BIO'
        for label in labels:
            if label[0] not in 'BIOES':
                raise ValueError(f'Unsupported label: {label}')
            if label[0] in 'ES':
                tag_scheme = 'BIOES'
        return tag_scheme

    @staticmethod
    def _gen_label_to_id_with_bio(labels: List[str], tag_scheme: str = 'BIOES') -> Dict[str, int]:
        label_to_id = {}
        if 'O' in tag_scheme:
            label_to_id['O'] = 0
        for label in labels:
            for tag in 'BIES':
                if tag in tag_scheme:
                    label_to_id[f'{tag}-{label}'] = len(label_to_id)
        return label_to_id

    @staticmethod
    def _spans_to_bio_labels(spans: List[Dict], length: int, tag_scheme: str = 'BIOES'):
        labels = [NON_ENTITY_LABEL] * length
        for span in spans:
            for i in range(span['start'], min(span['end'], length)):
                if span['type'] == PARTIAL_LABEL:
                    labels[i] = span['type']
                    continue
                if 'S' in tag_scheme and i == span['start'] == span['end'] - 1:
                    labels[i] = 'S-' + span['type']
                elif i == span['start']:
                    labels[i] = 'B-' + span['type']
                elif 'E' in tag_scheme and i == span['end'] - 1:
                    labels[i] = 'E-' + span['type']
                else:
                    labels[i] = 'I-' + span['type']
        return labels



# Copyright (c) Alibaba, Inc. and its affiliates.
# Fix some modelscope bugs temporarily.
# These bugs will be fixed in the next modelscope version.
import os

from modelscope.trainers.hooks.checkpoint_hook import BestCkptSaverHook, CheckpointHook
from modelscope.utils.checkpoint import save_checkpoint
from modelscope.utils.constant import LogKeys, ModelFile
from modelscope.utils.torch_utils import is_master


def _save_checkpoint(self, trainer):
    cur_save_name = self.save_file_name
    if cur_save_name is None:
        if self.by_epoch:
            cur_save_name = os.path.join(
                self.save_dir,
                f'best_{LogKeys.EPOCH}{trainer.epoch + 1}_{self.metric_key}{self._best_metric}.pth',
            )
        else:
            cur_save_name = os.path.join(
                self.save_dir,
                f'best_{LogKeys.ITER}{trainer.iter + 1}_{self.metric_key}{self._best_metric}.pth',
            )
    else:
        if '.' not in cur_save_name:
            cur_save_name = f'{cur_save_name}.pth'
        cur_save_name = os.path.join(self.save_dir, cur_save_name)

    meta = {
        'epoch': trainer.epoch,
        'iter': trainer.iter + 1,
        'inner_iter': trainer.inner_iter + 1,
        'rng_state': self.rng_state,
    }
    for i, hook in enumerate(trainer.hooks):
        meta[f'{hook.__class__}-{i}'] = hook.state_dict()

    if os.path.isfile(cur_save_name):
        os.remove(cur_save_name)
    save_checkpoint(trainer.model, cur_save_name, trainer.optimizer, trainer.lr_scheduler, meta)
    self._best_ckpt_file = cur_save_name
    self._save_pretrained(trainer)


def _save_pretrained(self, trainer):
    output_dir = os.path.join(self.save_dir, ModelFile.TRAIN_OUTPUT_DIR)
    from modelscope.trainers.parallel.utils import is_parallel

    if is_parallel(trainer.model):
        model = trainer.model.module
    else:
        model = trainer.model

    config = trainer.cfg.to_dict()
    # override pipeline by tasks name after finetune done,
    # avoid case like fill mask pipeline with a text cls task
    config['pipeline'] = {'type': config['task']}

    # remove parallel module that is not JSON serializable
    if 'parallel' in config and 'module' in config['parallel']:
        del config['parallel']['module']

    if hasattr(model, 'save_pretrained'):
        model.save_pretrained(
            output_dir,
            ModelFile.TORCH_MODEL_BIN_FILE,
            save_function=save_checkpoint,
            config=config,
            with_meta=False,
        )


def _should_save(self, trainer):
    return is_master() and self._is_best_metric(trainer.metric_values)


def after_run(self, trainer):  # noqa
    if self.restore_best:
        if is_master():
            self.load_checkpoint(self._best_ckpt_file, trainer)


def suppress_modelscope_ast_warning():  # noqa
    try:
        from modelscope.utils.logger import get_logger

        def filter_modelscope_ast_warning(record):
            return 'not found in ast index file' not in record.msg

        logger = get_logger('modelscope')
        logger.addFilter(filter_modelscope_ast_warning)
    except IsADirectoryError:
        pass


CheckpointHook._save_pretrained = _save_pretrained
BestCkptSaverHook._save_checkpoint = _save_checkpoint
BestCkptSaverHook._should_save = _should_save
BestCkptSaverHook.after_run = after_run

suppress_modelscope_ast_warning()



# Copyright (c) Alibaba, Inc. and its affiliates.
from .base import ENCODERS, Encoder, build_encoder
from .pytorch_rnn_encoder import GruEncoder, LstmEncoder, RnnEncoder
from .span_encoder import SpanEncoder



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Any, Dict, List, Union

from modelscope.preprocessors.builder import PREPROCESSORS
from modelscope.utils.constant import Fields

from adaseq.metainfo import Preprocessors

from ..constant import NONE_REL_LABEL, NONE_REL_LABEL_ID
from .nlp_preprocessor import NLPPreprocessor


@PREPROCESSORS.register_module(
    Fields.nlp, module_name=Preprocessors.relation_extraction_preprocessor
)
class RelationExtractionPreprocessor(NLPPreprocessor):
    """Relation Extraction data preprocessor"""

    def __init__(self, model_dir: str, labels: List[str], **kwargs):
        label_to_id = self._gen_label2id(labels)
        super().__init__(model_dir, label_to_id=label_to_id, return_offsets=True, **kwargs)

    def __call__(self, data: Union[str, List, Dict]) -> Dict[str, Any]:
        """prepare inputs for Relation Extraction model."""
        output = super().__call__(data)

        if isinstance(data, Dict) and 'label' in data:
            output['label_id'] = self.label_to_id[data['label']]
        output['so_head_mask'] = data['so_head_mask']
        return output

    @staticmethod
    def _gen_label2id(labels: List[str]) -> Dict[str, int]:
        label2id = {}
        label2id[NONE_REL_LABEL] = NONE_REL_LABEL_ID
        for label in labels:
            if label != NONE_REL_LABEL:
                label2id[f'{label}'] = len(label2id)
        return label2id



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Dict

from modelscope.metrics.base import Metric
from modelscope.metrics.builder import METRICS, MetricKeys
from modelscope.utils.tensor_utils import torch_nested_detach, torch_nested_numpify
from seqeval.metrics import accuracy_score, classification_report

from adaseq.data.constant import PAD_LABEL_ID
from adaseq.metainfo import Metrics


@METRICS.register_module(module_name=Metrics.sequence_labeling_metric)
@METRICS.register_module(module_name=Metrics.ner_metric)
class SequenceLabelingMetric(Metric):
    """The metric computation class for sequence-labeling tasks.

    This metric class uses seqeval to calculate scores.

    Args:
        return_macro_f1 (bool, *optional*):
            Whether to return macro-f1, default False.
        return_class_level_metric (bool, *optional*):
            Whether to return every class's detailed metrics, default False.
    """

    def __init__(
        self, return_macro_f1=False, return_class_level_metric=False, mode=None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.return_macro_f1 = return_macro_f1
        self.return_class_level_metric = return_class_level_metric
        self.mode = mode
        self.preds = []
        self.golds = []

    def add(self, outputs: Dict, inputs: Dict):
        """Collect batch outputs"""
        pred_results = outputs['predicts']
        ground_truths = inputs['label_ids']
        self.preds.extend(torch_nested_numpify(torch_nested_detach(pred_results)).tolist())
        self.golds.extend(torch_nested_numpify(torch_nested_detach(ground_truths)).tolist())

    def evaluate(self):
        """Calculate metrics, returning precision, recall, f1-score, accuracy in a dictionary

        Returns:
            scores (Dict):
                precision (float): micro averaged precision
                recall (float): micro averaged recall
                f1 (float): micro averaged f1-score
                accuracy (float): micro averaged accuracy
                tp (Dict): metrics for each label
                    precision (float): precision of each label
                    recall (float): recall of each label
                    f1 (float): f1 of each label
                    support (int): the number of occurrences of each label in ground truth
        """
        id2label = self.trainer.train_preprocessor.id_to_label

        pred_labels = [
            [id2label[p] for p, g in zip(pred, gold) if g != PAD_LABEL_ID]
            for pred, gold in zip(self.preds, self.golds)
        ]
        gold_labels = [[id2label[g] for g in gold if g != PAD_LABEL_ID] for gold in self.golds]

        report = classification_report(gold_labels, pred_labels, output_dict=True, mode=self.mode)

        report.pop('weighted avg')
        macro_score = report.pop('macro avg')
        micro_score = report.pop('micro avg')

        scores = {}
        scores[MetricKeys.PRECISION] = micro_score['precision']
        scores[MetricKeys.RECALL] = micro_score['recall']
        scores[MetricKeys.F1] = micro_score['f1-score']
        scores[MetricKeys.ACCURACY] = accuracy_score(y_true=gold_labels, y_pred=pred_labels)
        if self.return_macro_f1:
            scores['macro-f1'] = macro_score['f1-score']
        if self.return_class_level_metric:
            for tp, tp_score in report.items():
                scores[tp] = {}
                scores[tp][MetricKeys.PRECISION] = round(tp_score['precision'], 4)
                scores[tp][MetricKeys.RECALL] = round(tp_score['recall'], 4)
                scores[tp][MetricKeys.F1] = round(tp_score['f1-score'], 4)
                scores[tp]['support'] = tp_score['support']
        return scores



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Any, Dict

from adaseq.data.constant import PAD_LABEL_ID
from adaseq.metainfo import DataCollators

from .base import DATA_COLLATORS, DataCollatorWithPadding


@DATA_COLLATORS.register_module(module_name=DataCollators.sequence_labeling_data_collator)
class SequenceLabelingDataCollatorWithPadding(DataCollatorWithPadding):
    """Collator for the sequence labeling task"""

    pad_label_id: int = PAD_LABEL_ID

    def padding(self, batch: Dict[str, Any], padding_side: str, **kwargs) -> Dict[str, Any]:
        """pad label sequence `label_ids`"""
        return super().padding(batch, padding_side, 'label_ids', self.pad_label_id)



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Any, Dict, Optional, Union

import torch
import torch.nn as nn
from modelscope.models.builder import MODELS
from torch.nn import BCELoss, BCEWithLogitsLoss

from adaseq.metainfo import Models, Tasks
from adaseq.models.base import Model
from adaseq.modules.decoders import Decoder, PairwiseCRF
from adaseq.modules.dropouts import WordDropout
from adaseq.modules.embedders import Embedder
from adaseq.modules.encoders import SpanEncoder


class WBCEWithLogitsLoss:
    """Weighed BCE loss, multiply the loss of positive examples with a scaler"""

    def __init__(self, pos_weight=1.0):
        self.pos_weight = pos_weight
        self.loss = BCEWithLogitsLoss(reduction='none')

    def __call__(self, y_pred, y_true):  # noqa: D102
        loss = self.loss(y_pred, y_true.float())

        if self.pos_weight != 1:
            weights_pr = torch.ones_like(y_true).float()  # B x C
            weights_pr[y_true > 0] = self.pos_weight
            loss = (loss * weights_pr).mean()
        else:
            loss = loss.mean()

        return loss


class WBCEWithLogitsLossUFET:
    """Weighed BCE loss, multiply the loss of positive examples with a scaler,
    Apply the trick in Ultra-Fine Entity Typing with Weak Supervision from a Masked Language Model"""

    def __init__(self, pos_weight=1.0):
        self.pos_weight = pos_weight
        self.loss = BCEWithLogitsLoss(reduction='none')

    def __call__(self, y_pred, y_true):  # noqa: D102
        loss = self.loss(y_pred, y_true.float())
        B, L = y_true.shape
        coarse = y_true[:, :9].sum(-1) > 0
        fine = y_true[:, 9:130].sum(-1) > 0
        ultra = y_true[:, 130:].sum(-1) > 0
        a = coarse.view(-1, 1) * torch.ones_like(y_true[:, :9])
        b = fine.view(-1, 1) * torch.ones_like(y_true[:, 9:130])
        c = ultra.view(-1, 1) * torch.ones_like(y_true[:, 130:])
        weight_cfu = torch.cat([a, b, c], -1)
        loss = loss * weight_cfu
        if self.pos_weight != 1:
            weights_pr = torch.ones_like(y_true).float()  # B x C
            weights_pr[y_true > 0] = self.pos_weight
            loss = (loss * weights_pr).mean()
        else:
            loss = loss.mean()

        return loss


@MODELS.register_module(Tasks.entity_typing, module_name=Models.multilabel_span_typing_model)
class MultiLabelSpanTypingModel(Model):
    """Span based MultiLabel Entity Typing model

    This model is used for multilabel entity typing tasks.

    Args:
        num_labels (int): number of labels
        embedder (Union[Embedder, str], `optional`): embedder used in the model.
            It can be an `Embedder` instance or an embedder config file or an embedder config dict.
        span_encoder_method (str): concat
        word_dropout (float, `optional`): word dropout rate, default `0.0`.
        loss_function (str, `optional'): loss function, default 'BCE',
        class_threshold (float, `optional`): classification threshold default. '0.5'
        use_biaffine (bool, `optional`): whether to use biaffine for span representation, default `False`.
        **kwargs: other arguments
    """

    def __init__(
        self,
        id_to_label: Dict[int, str],
        embedder: Union[Embedder, str],
        span_encoder_method: str = 'concat',
        dropout: float = 0.0,
        word_dropout: bool = False,
        loss_function: str = 'BCE',
        class_threshold: float = 0.5,
        use_biaffine: bool = False,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)
        if isinstance(embedder, Embedder):
            self.embedder = embedder
        else:
            self.embedder = Embedder.from_config(embedder)

        self.span_encoder = SpanEncoder(
            self.embedder.get_output_dim(), span_encoder_method, **kwargs
        )
        self.linear_input_dim = self.span_encoder.output_dim

        self.id_to_label = id_to_label
        self.num_labels = len(id_to_label)
        self.linear = nn.Linear(self.linear_input_dim, self.num_labels)

        self.use_dropout = dropout > 0.0
        if self.use_dropout:
            if word_dropout:
                self.dropout = WordDropout(dropout)
            else:
                self.dropout = nn.Dropout(dropout)

        self.loss_function_type = loss_function
        self.class_threshold = class_threshold
        self.use_biaffine = use_biaffine

        if self.loss_function_type == 'BCE':
            self.sigmoid = nn.Sigmoid()
            self.loss_fn = BCELoss()

    def forward(  # noqa: D102
        self,
        tokens: Dict[str, Any],
        mention_boundary: torch.LongTensor,
        mention_mask: torch.LongTensor,
        type_ids: Optional[torch.LongTensor] = None,
        meta: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:  # TODO docstring
        # B x M -> B*M x 1
        logits = self._forward(tokens, mention_boundary)
        mask = mention_mask.reshape(-1).unsqueeze(-1)
        if self.training:
            loss = self._calculate_loss(logits, type_ids, mask)
            outputs = {'logits': logits, 'loss': loss}
        else:
            batch_size, max_mention_per_sent = type_ids.shape[0:2]  # TODO: no supervision
            logits = logits.reshape(batch_size, max_mention_per_sent, -1)
            predicts = self.classify(logits, mask)
            outputs = {'logits': logits, 'predicts': predicts}
        return outputs

    def _forward(self, tokens: Dict[str, Any], mention_boundary: torch.LongTensor) -> torch.Tensor:
        # B*M x K
        if self.use_biaffine:
            span_start_reprs, span_end_reprs = self._token2span_encode(tokens, mention_boundary)
            span_reprs = span_start_reprs * span_end_reprs
        else:
            span_reprs = self._token2span_encode(tokens, mention_boundary)

        # B*M x label_num
        logits = self.linear(span_reprs)
        return logits

    def _token2span_encode(
        self, tokens: Dict[str, Any], mention_boundary: torch.LongTensor
    ) -> torch.Tensor:
        embed = self.embedder(**tokens)
        # embed: B x W x K
        if self.use_dropout:
            embed = self.dropout(embed)
        # span_reprs: B x S x D: trans sentence sequence to mentions sequence
        # the shape of boundary_ids is batch_size x 2 x max_mention_num, providing the detected mentions
        span_reprs = self.span_encoder(embed, mention_boundary)
        return span_reprs

    def _calculate_loss(self, logits, targets, mask):
        logits = logits * mask  # B*M x L
        if self.loss_function_type == 'BCE':
            logits = self.sigmoid(logits)  # B*M x L
            targets = targets.reshape(-1, self.num_labels).to(torch.float32)  # B*M x L
            loss = self.loss_fn(logits, targets) * mask
        else:
            raise ValueError('Unsupported loss %s', self.loss_function_type)
        loss = loss.sum() / (mask.sum())
        return loss

    def classify(self, logits, mention_mask):  # noqa: D102
        if self.loss_function_type == 'BCE':
            logits = self.sigmoid(logits)  # B*M x L
            predicts = torch.where(logits > self.class_threshold, 1, 0)  # B*M x L
        else:
            raise ValueError('Unsupported loss %s', self.loss_function_type)

        predicts = predicts.detach().cpu().numpy()
        batch_mention_mask = mention_mask.detach().cpu().numpy()
        batch_mentions = []
        for prediction, mask in zip(predicts, batch_mention_mask):
            mentions = []
            for pred, flag in zip(prediction, mask):
                if flag == 1:
                    types = set(self.id_to_label[i] for i, p in enumerate(pred) if p == 1)
                    mentions.append(types)
            batch_mentions.append(mentions)

        return batch_mentions


@MODELS.register_module(Tasks.entity_typing, module_name=Models.multilabel_concat_typing_model)
class MultiLabelConcatTypingModel(Model):
    """Concat based Single Mention MultiLabel Entity Typing model

    This model is used for single mention multilabel entity typing tasks.
    Input format [cls] sentence [sep] mention [sep]

    Args:
        labels (List): list of labels
        embedder (Union[Embedder, str], `optional`): embedder used in the model.
            It can be an `Embedder` instance or an embedder config file or an embedder config dict.
        word_dropout (float, `optional`): word dropout rate, default `0.0`.
        loss_function (str, `optional'): loss function, default 'BCE', other options `WBCE`, `WBCE-UFET'
        class_threshold (float, `optional`): classification threshold default. '0.5'
        pos_weight (float, `optional`): multiplier on loss of positive examples, default 1.0.
        decoder (Union[Encoder, str], `optional`), can be linear or pairwise-crf
        **kwargs: other arguments
    """

    def __init__(
        self,
        id_to_label: Dict[int, str],
        embedder: Union[Embedder, Dict] = None,
        dropout: float = 0.0,
        word_dropout: bool = False,
        loss_function: str = 'BCE',
        class_threshold: float = 0.5,
        pos_weight: float = 1.0,
        decoder: Union[Decoder, str] = None,
        **kwargs
    ) -> None:
        super().__init__()
        self.id_to_label = id_to_label
        self.num_labels = len(id_to_label)
        if isinstance(embedder, Embedder):
            self.embedder = embedder
        else:
            if isinstance(embedder, dict):
                embedder['drop_special_tokens'] = False
            self.embedder = Embedder.from_config(embedder)

        self.linear_input_dim = self.embedder.get_output_dim()

        self.use_dropout = dropout > 0.0
        if self.use_dropout:
            if word_dropout:
                self.dropout = WordDropout(dropout)
            else:
                self.dropout = nn.Dropout(dropout)

        self.loss_function_type = loss_function
        self.class_threshold = class_threshold
        self.pos_weight = pos_weight

        if self.loss_function_type == 'BCE':
            self.sigmoid = nn.Sigmoid()
            self.loss_fn = BCEWithLogitsLoss()
        elif self.loss_function_type == 'WBCE':
            self.sigmoid = nn.Sigmoid()
            self.loss_fn = WBCEWithLogitsLoss(pos_weight=self.pos_weight)
        elif self.loss_function_type == 'WBCE-UFET':
            self.sigmoid = nn.Sigmoid()
            self.loss_fn = WBCEWithLogitsLossUFET(pos_weight=self.pos_weight)

        self.linear = nn.Linear(self.linear_input_dim, self.num_labels)

        self.decoder_type = decoder.pop('type')
        assert self.decoder_type in [
            'pairwise-crf',
            'linear',
        ], 'decoder_type {} unimplemented'.format(self.decoder_type)

        if self.decoder_type == 'pairwise-crf':
            self.decoder = PairwiseCRF(
                labels=[id_to_label[i] for i in range(self.num_labels)], **decoder
            )

    def forward(  # noqa: D102
        self,
        tokens: Dict[str, Any],
        type_ids: Optional[torch.LongTensor] = None,
        meta: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:  # TODO docstring
        # B x M -> B*M x 1
        logits = self._forward(tokens)
        if self.training and type_ids is not None:
            loss = self.loss_fn(logits, type_ids.float())
            outputs = {'logits': logits, 'loss': loss}
        else:
            predicts = self.classify(logits)
            outputs = {'logits': logits, 'predicts': predicts}
        return outputs

    def _forward(self, tokens: Dict[str, Any]) -> torch.Tensor:
        span_reprs = self._token2span_encode(tokens)  # B x K
        # B*M x label_num
        logits = self.linear(span_reprs)
        if self.decoder_type == 'pairwise-crf':
            logits = self.decoder(logits)

        return logits

    def _token2span_encode(self, tokens: Dict[str, Any]) -> torch.Tensor:
        embed = self.embedder(**tokens)
        # embed: B x W x K
        if self.use_dropout:
            embed = self.dropout(embed)
        # use the cls embedding
        return embed[:, 0, :]

    def classify(self, logits):  # noqa: D102
        logits = self.sigmoid(logits)  # B x L
        predicts = []
        for i in range(len(logits)):
            pred_ = (logits[i] > 0.5).nonzero()  # at least predict one
            if len(pred_) < 1:
                pred_ = logits[i].topk(k=1).indices
            types = set(self.id_to_label[i] for i in pred_.view(-1).cpu().numpy())
            predicts.append([types])
        return predicts



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Any, Dict, List, Optional, Union

import torch
import torch.nn as nn
from modelscope.models.builder import MODELS

from adaseq.metainfo import Models, Tasks
from adaseq.models.base import Model
from adaseq.modules.biaffine import Biaffine
from adaseq.modules.dropouts import WordDropout
from adaseq.modules.embedders import Embedder
from adaseq.modules.encoders import Encoder
from adaseq.modules.util import get_tokens_mask


@MODELS.register_module(Tasks.named_entity_recognition, module_name=Models.biaffine_ner_model)
class BiaffineNerModel(Model):
    """
    Named Entity Recognition as Dependency Parsing (Yu et al., ACL 2020)
    ref: https://aclanthology.org/2020.acl-main.577/
    """

    def __init__(
        self,
        id_to_label: Union[Dict[int, str], List[str]],
        embedder: Optional[Union[Embedder, Dict]] = None,
        encoder: Optional[Union[Encoder, Dict[str, Any]]] = None,
        biaffine_ffnn_size: int = -1,
        biaffine_bias: bool = True,
        flat_ner: bool = True,
        dropout: float = 0.0,
        word_dropout: bool = False,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.id_to_label = id_to_label
        self.num_labels = len(id_to_label) + 1  # leave 0 as non-entity label
        self.flat_ner = flat_ner

        if isinstance(embedder, Embedder):
            self.embedder = embedder
        else:
            self.embedder = Embedder.from_config(embedder, **kwargs)
        hidden_size = self.embedder.get_output_dim()

        if encoder is None:
            self.encoder = None
        else:
            if isinstance(encoder, Encoder):
                self.encoder = encoder
            else:
                self.encoder = Encoder.from_config(encoder)
            assert hidden_size == self.encoder.get_input_dim()
            hidden_size = self.encoder.get_output_dim()

        self.scorer = BiaffineScorer(
            hidden_size, biaffine_ffnn_size, self.num_labels, biaffine_bias
        )

        self.use_dropout = dropout > 0.0
        if self.use_dropout:
            if word_dropout:
                self.dropout = WordDropout(dropout)
            else:
                self.dropout = nn.Dropout(dropout)

    def forward(
        self,
        tokens: Dict[str, Any],
        span_labels: Optional[torch.LongTensor] = None,
        meta: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """
        Args:
        tokens (Dict[str, Any]):
            embedder inputs.
        span_labels (torch.LongTensor):
            span labels with size (len, len), span_labels[start][end] == label_id
        meta (Dict[str, Any]):
            meta data of raw inputs.
        """
        x = self.embedder(**tokens)
        mask = get_tokens_mask(tokens, x.size(1))

        if self.use_dropout:
            x = self.dropout(x)

        if self.encoder is not None:
            x = self.encoder(x, mask)

            if self.use_dropout:
                x = self.dropout(x)

        span_scores = self.scorer(x)
        outputs = {'span_scores': span_scores}

        if self.training and span_labels is not None:
            loss = self._calculate_loss(outputs['span_scores'], span_labels, mask)
            outputs.update(loss=loss)
        else:
            predicts = self.decode(outputs['span_scores'], mask)
            outputs.update(predicts=predicts)
        return outputs

    def _calculate_loss(
        self, span_scores: torch.Tensor, span_labels: torch.LongTensor, mask: torch.BoolTensor
    ) -> torch.Tensor:
        """
        span_labels : (batch_size, seq_len, seq_len)
        span_scores : (batch_size, seq_len, seq_len, num_classes)
        """
        label_mask = torch.triu(mask.unsqueeze(-1).expand_as(span_labels).clone())
        loss = nn.functional.cross_entropy(
            span_scores.reshape(-1, self.num_labels),
            span_labels.masked_fill(~label_mask, -100).reshape(-1),
        )
        return loss

    def decode(self, span_scores, mask):
        """
        :param span_scores: (b, t, t, c)
        :param mask: (b, t)
        :return:
        """
        lengths = mask.sum(-1).tolist()
        # (b, t, t)
        type_idxs = span_scores.detach().argmax(dim=-1)
        # (b, t, t)
        span_max_score = (
            span_scores.detach().gather(dim=-1, index=type_idxs.unsqueeze(-1)).squeeze(-1)
        )
        final = []
        for span_score, tids, l in zip(span_max_score, type_idxs, lengths):
            cands = []
            for s in range(l):
                for e in range(s, l):
                    type_id = tids[s, e].item()
                    if type_id > 0:
                        # we do not have non-entity label, we set index 0 as non-entity label
                        # so we substract 1 on type_id
                        cands.append((s, e, type_id - 1, span_score[s, e].item()))

            pre_res = []
            for s, e, cls, _ in sorted(cands, key=lambda x: x[3], reverse=True):
                for pred_s, pred_e, _ in pre_res:
                    # contraint on the token level, so convert sub-token indices to token
                    if s < pred_s <= e < pred_e or pred_s < s <= pred_e < e:
                        break  # for both nested and flat ner no clash is allowed
                    if self.flat_ner and (s <= pred_s <= pred_e <= e or pred_s <= s <= e <= pred_e):
                        break  # for flat ner nested mentions are not allowed
                else:
                    pre_res.append((s, e + 1, self.id_to_label[cls]))
            final.append(pre_res)

        return final


class BiaffineScorer(nn.Module):
    """Biaffine scorer."""

    def __init__(
        self,
        input_size: int,
        ffnn_size: int,
        num_cls: int,
        # ffnn_drop: float = 0.33,
        bias: bool = True,
    ) -> None:
        super().__init__()
        self._act = nn.ELU()
        self.mlp_start = nn.Linear(input_size, ffnn_size)
        self.mlp_end = nn.Linear(input_size, ffnn_size)
        self.span_biaff = Biaffine(ffnn_size, num_cls, bias=(bias, bias))
        # self._dropout = WordDropout(ffnn_drop)

    def forward(self, enc_hn: torch.Tensor) -> torch.Tensor:
        """biaffine attention scores."""
        start_feat = self._act(self.mlp_start(enc_hn))
        end_feat = self._act(self.mlp_end(enc_hn))

        # if self.training:
        #     start_feat = self._dropout(start_feat)
        #     end_feat = self._dropout(end_feat)

        # (bz, len, len, num_lbl)
        span_score = self.span_biaff(start_feat, end_feat)
        return span_score



# Copyright (c) Alibaba, Inc. and its affiliates.
import logging
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from modelscope.preprocessors.base import Preprocessor
from modelscope.preprocessors.builder import PREPROCESSORS
from modelscope.preprocessors.builder import build_preprocessor as ms_build_preprocessor
from modelscope.utils.config import ConfigDict
from modelscope.utils.constant import Fields, ModeKeys

from adaseq.data.tokenizer import build_tokenizer
from adaseq.metainfo import Preprocessors

logger = logging.getLogger(__name__)


@PREPROCESSORS.register_module(Fields.nlp, module_name=Preprocessors.nlp_preprocessor)
class NLPPreprocessor(Preprocessor):
    """
    Some common pre-process operations for NLP tasks.

    Args:
        model_dir (str): pre-trained model name or path.
        is_word2vec (bool): if True, tokens will be indexed in word2vec style,
            i.e., 1 to 1 mapping, without word pieces.
        tokenizer_kwargs (Optional[Dict[str, Any]]): some arguments to init tokenizer
            from huggingface, modelscope or ...
        max_length (int): we will discard tokens that exceed the `max_length`.
            So please take care of this argument.
        return_offsets (bool): if `True`, compute sub-token offset mapping for the
            original sequence reconstruction in the `TransformerEncoder`.
        add_special_tokens (bool): add special tokens of pre-trained models to the
            input, it is only effective when `return_offsets==False`.
        labels (List[str]): a label list, which is used to setup a `label_to_id`
            mappging if the following `label_to_id` was not provided.
        label_to_id (Dict[str, int]): a dict maps label to index,
            such as `{'O': 0, 'B-LOC': 1, 'I-LOC': 2}`.
        return_original_view (bool): if `True`, return token_ids and other tensors
            that without padded context, only used in retrieval-augmented models.
    """

    def __init__(
        self,
        model_dir: str,
        is_word2vec: bool = False,
        tokenizer_kwargs: Optional[Dict[str, Any]] = None,
        max_length: int = 512,
        return_offsets: bool = False,
        add_special_tokens: bool = True,
        labels: Optional[List[str]] = None,
        label_to_id: Optional[Dict[str, int]] = None,
        return_original_view: Optional[bool] = None,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)
        if 'word2vec' in model_dir and is_word2vec is False:
            is_word2vec = True
            logger.warn('You are using word2vec embedder, auto set `is_word2vec = True`')

        self.is_word2vec = is_word2vec
        self.max_length = max_length
        self.add_special_tokens = add_special_tokens
        self.return_offsets = return_offsets
        self.return_original_view = return_original_view
        self.tokenizer = build_tokenizer(
            model_dir, is_word2vec=is_word2vec, **(tokenizer_kwargs or {})
        )

        if label_to_id is not None:
            self.label_to_id = label_to_id
        elif labels is not None:
            self.label_to_id = self.make_label_to_id(labels)
        else:
            raise ValueError('Must have one of `labels` or `label_to_id`')
        self.id_to_label = {v: k for k, v in sorted(self.label_to_id.items(), key=lambda x: x[1])}
        # make sure they are aligned.
        assert len(self.id_to_label) not in self.id_to_label
        assert len(self.id_to_label) - 1 in self.id_to_label

    def __call__(self, data: Union[str, List, Dict]) -> Dict[str, Any]:
        """
        Encode one instance, it could be a text str, a list of tokens for a dict.

        Returns:
            Dict[str, Any]: `{'tokens': tokenized and encoded tensors, 'meta': data input}`
        """
        if isinstance(data, str):
            data = {'text': data}
        if isinstance(data, List):
            data = {'tokens': data}

        output_dict = {'meta': data}

        if 'tokens' in data:
            output_dict['tokens'] = self.encode_tokens(data['tokens'])
        elif 'text' in data:
            output_dict['tokens'] = self.encode_text(data['text'])
        else:
            raise ValueError('Data sample must have "text" or "tokens" field!')

        if self.return_original_view is not None:
            output_dict['origin_mask'] = data['mask']
            if self.return_original_view:
                # return token_ids and other tensors that without padded context,
                # only used in retrieval-augmented models.
                output_dict['origin_tokens'] = self.encode_tokens_origin_view(data)

        if self.mode == ModeKeys.INFERENCE:
            for key, value in output_dict['tokens'].items():
                output_dict['tokens'][key] = np.expand_dims(np.array(value), 0)

        return output_dict

    def encode_text(self, text: str) -> Dict[str, Any]:
        """encode `text` to ids"""
        inputs_with_offsets = self.tokenizer(
            text, add_special_tokens=False, return_offsets_mapping=self.return_offsets
        )
        encoded = self.encode_tokens(inputs_with_offsets.tokens())
        if self.return_offsets and self.mode == ModeKeys.INFERENCE:
            encoded['offset_mapping'] = inputs_with_offsets['offset_mapping']
        return encoded

    def encode_tokens(self, tokens: List[str]) -> Dict[str, Any]:
        """
        Convert tokens to ids, add some mask.
        """
        if self.is_word2vec:
            return self.encode_tokens_word2vec(tokens)
        else:
            return self.encode_tokens_wordpiece(tokens)

    def encode_tokens_word2vec(self, tokens: List[str]) -> Dict[str, Any]:
        """
        Convert tokens to ids, one by one via vocab, no word pieces.
        """
        input_ids = self.tokenizer.convert_tokens_to_ids(tokens)
        encoded = {'input_ids': input_ids, 'mask': [True] * len(tokens)}
        return encoded

    def encode_tokens_wordpiece(self, tokens: List[str]) -> Dict[str, Any]:
        """
        Convert tokens to ids by word piece tokenizer.
        """
        input_ids = []
        # the corresponding inclusive sub-token span of tokens
        offsets: List[Optional[Tuple[int, int]]] = []

        max_length = self.max_length

        if self.add_special_tokens:
            input_ids.append(self.tokenizer.cls_token_id)
            offsets.append((0, 0))

            # if `add_special_tokens`, the max_length should minus 1 for the appending `[SEP]`
            max_length -= int(self.add_special_tokens)

        for token_string in tokens:
            wordpieces = self.tokenizer.encode_plus(
                token_string,
                add_special_tokens=False,
                return_tensors=None,
                return_offsets_mapping=False,
                return_attention_mask=False,
            )
            wp_ids = wordpieces['input_ids']

            # For tokens that don't correspond to any word pieces, we set it to [UNK].
            if len(wp_ids) == 0:
                wp_ids = [self.tokenizer.unk_token_id]

            offsets.append((len(input_ids), len(input_ids) + len(wp_ids) - 1))
            input_ids.extend(wp_ids)

            if len(input_ids) >= max_length:
                # discard sub-tokens that exceed the `max_length`
                input_ids = input_ids[:max_length]
                offsets[-1] = (offsets[-1][0], len(input_ids) - 1)
                break

        if self.add_special_tokens:
            offsets.append((len(input_ids), len(input_ids)))
            input_ids.append(self.tokenizer.sep_token_id)

        encoded = {
            'input_ids': input_ids,
            'attention_mask': [True] * len(input_ids),
            'has_special_tokens': self.add_special_tokens,
        }
        if self.return_offsets:
            encoded['mask'] = [True] * len(offsets)
            encoded['offsets'] = offsets
        return encoded

    def make_label_to_id(self, labels: List[str]) -> Dict[str, int]:
        """
        Generate `label_to_id` mapping.
        You can override this method to customize the mapping.
        NOTE: The `self.labels` could be modified in this method.
        """
        return {label: i for i, label in enumerate(labels)}

    def encode_tokens_origin_view(self, data: Dict[str, Any]) -> Dict[str, Any]:
        """encode tokens when using retrieval-augmented multi-view model."""
        tokens = data['tokens']
        mask = data.get('mask', [True] * len(tokens))
        # remove the padded context
        origin_length = sum(mask)
        origin_tokens = tokens[:origin_length]
        return self.encode_tokens(origin_tokens)


def build_preprocessor(config: ConfigDict, **kwargs) -> Preprocessor:
    """Build preprocessor from config"""
    # get `field_name` for loading modelscope preprocessors,
    # Not sure who will need this.
    field_name = config.get('field_name', Fields.nlp)
    return ms_build_preprocessor(config, field_name, kwargs)  # type: ignore



# Copyright (c) Alibaba, Inc. and its affiliates.
import json
import logging
import os

from datasets.arrow_dataset import Dataset
from modelscope.msdatasets.task_datasets.torch_base_dataset import TorchTaskDataset
from modelscope.preprocessors.base import Preprocessor
from modelscope.trainers.builder import TRAINERS
from modelscope.trainers.builder import build_trainer as ms_build_trainer
from modelscope.trainers.trainer import EpochBasedTrainer
from modelscope.utils.config import Config, ConfigDict
from modelscope.utils.constant import ModeKeys
from torch import nn

from adaseq.data.data_collators.base import DataCollatorWithPadding
from adaseq.data.dataset_manager import DatasetManager
from adaseq.metainfo import Trainers
from adaseq.models.base import Model

from .default_config import DEFAULT_CONFIG
from .lr_scheduler import build_lr_scheduler
from .optimizer import build_optimizer

logger = logging.getLogger(__name__)


@TRAINERS.register_module(module_name=Trainers.default_trainer)
class DefaultTrainer(EpochBasedTrainer):
    """Default trainer class for AdaSeq.

    This trainer inherits from EpochBasedTrainer with some modifications.
    It implements some common data processing functions which are convenient for training a model.
    It also implements a basic test function for evaluate a trained model on the test dataset.

    Args:

    cfg_file (`str`): required
        The path of `Config` of this trial.
    work_dir (`str`): required
        The created directionary to save all produced files in training.
    dataset_manager (`DatasetManager`): required
        A `DatasetManager` with already loaded datasets.
    data_collator (`DataCollatorWithPadding`): required
        A `DataCollatorWithPadding` object to collect data batches.
    preprocessor (`Preprocessor`): required
        A `Preprocessor` to process data before training.
    seed (`int`): required
        Random seed.
    device (`str`): 'cpu' by default
        A device string, e.g., 'cpu', 'cuda', 'cuda:1'.
    kwargs (`dict`):
        Other keyword arguments for `EpochBasedTrainer` of `modelscope`.
    """

    def __init__(
        self,
        cfg_file: str,
        work_dir: str,
        dataset_manager: DatasetManager,
        data_collator: DataCollatorWithPadding,
        preprocessor: Preprocessor,
        seed: int = 42,
        device: str = 'cpu',
        **kwargs,
    ) -> None:
        super().__init__(
            model=None,
            cfg_file=cfg_file,
            cfg_modify_fn=None,
            data_collator=data_collator,
            train_dataset=dataset_manager.train,
            eval_dataset=dataset_manager.valid,
            preprocessor=preprocessor,
            work_dir=work_dir,
            seed=seed,
            device=device,
            **kwargs,
        )

        # Setup testset if there is one
        if dataset_manager.test is not None:
            self.test_dataset = self.to_task_dataset(
                dataset_manager.test,
                mode=ModeKeys.EVAL,
                preprocessor=self.eval_preprocessor,
            )
        logger.info(f'device: {self.device}')

    def rebuild_config(self, config: Config) -> Config:
        """
        Override this func to add adaseq default config.
        """
        config = Config.from_file(config.filename)
        config.merge_from_dict(DEFAULT_CONFIG, force=False)
        return config

    def build_model(self) -> nn.Module:
        """
        Override this func to build adaseq `Model`.
        """
        return Model.from_config(self.cfg)

    def build_optimizer(self, cfg: ConfigDict, default_args: dict = None):
        """
        Override this func to customize the optimizer.
        """
        return build_optimizer(self.model, cfg, default_args)

    def build_lr_scheduler(self, cfg: ConfigDict, default_args: dict = None):
        """
        Override this func to customize the lr_scheduler.
        """
        batch_size = self.cfg.train.dataloader.batch_size_per_gpu
        total_steps = max(len(self.train_dataset) / batch_size, 1) * self._max_epochs
        return build_lr_scheduler(cfg, int(total_steps), default_args)

    def to_task_dataset(
        self, dataset: Dataset, mode: str, preprocessor: Preprocessor, **kwargs
    ) -> TorchTaskDataset:
        """
        Override this func to build task dataset from only `datasets.Dataset`.
        """
        task_dataset = TorchTaskDataset(
            dataset, mode=mode, preprocessor=preprocessor, type=self.cfg.model.type
        )
        task_dataset.trainer = self
        return task_dataset

    def _init_file_logger(self):
        from modelscope.trainers.hooks.logger.text_logger_hook import TextLoggerHook

        self._file_logger = None
        for hook in self.hooks:
            if isinstance(hook, TextLoggerHook):
                self._file_logger = hook
                break

    def dump_log(self, log_dict):
        """Dump dict to log file"""
        if self._file_logger is not None:
            self._file_logger._dump_log(log_dict)

    def train(self, checkpoint_path=None, *args, **kwargs):
        """Train a model with training set"""
        return super().train(checkpoint_path=checkpoint_path, *args, **kwargs)

    def test(self, checkpoint_path=None):
        """Evaluate a trained model on testing set"""
        backup_eval_dataset = self.eval_dataset
        self.eval_dataset = self.test_dataset
        metric_values = self.evaluate(checkpoint_path)
        self.eval_dataset = backup_eval_dataset

        # log to terminal
        metric_string = json.dumps(metric_values, indent=2, ensure_ascii=False)
        logger.info('test: ' + metric_string)

        # log to file
        self._init_file_logger()
        from collections import OrderedDict

        log_dict = OrderedDict(mode='test', seed=self._seed, **metric_values)
        self.dump_log(log_dict)

        return metric_values


def build_trainer(name: str, **kwargs) -> EpochBasedTrainer:
    """build trainer from config"""
    if 'LOCAL_RANK' not in os.environ:
        os.environ['LOCAL_RANK'] = kwargs.get('local_rank', '0')
    if 'WORLD_SIZE' in os.environ and int(os.environ['WORLD_SIZE']) > 1:
        kwargs.update(launcher='pytorch', device='gpu')

    trainer = ms_build_trainer(name, kwargs)
    return trainer



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Any, Dict, List, Optional, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from modelscope.models.builder import MODELS
from modelscope.utils.config import ConfigDict

from adaseq.data.constant import PAD_LABEL_ID
from adaseq.metainfo import Models, Tasks
from adaseq.models.base import Model
from adaseq.modules.decoders import CRF, PartialCRF
from adaseq.modules.dropouts import WordDropout
from adaseq.modules.embedders import Embedder
from adaseq.modules.encoders import Encoder
from adaseq.modules.util import get_tokens_mask


@MODELS.register_module(Tasks.word_segmentation, module_name=Models.sequence_labeling_model)
@MODELS.register_module(Tasks.part_of_speech, module_name=Models.sequence_labeling_model)
@MODELS.register_module(Tasks.named_entity_recognition, module_name=Models.sequence_labeling_model)
class SequenceLabelingModel(Model):
    """Sequence labeling model

    This model is used for sequence labeling tasks.
    Various decoders are supported, including argmax, crf, partial-crf, etc.

    Args:
        num_labels (int): number of labels
        embedder (Union[Embedder, str], `optional`): embedder used in the model.
            It can be an `Embedder` instance or an embedder config file or an embedder config dict.
        dropout (float, `optional`): dropout rate, default `0.0`.
        word_dropout (bool): if `True`, use `WordDropout`.
        use_crf (bool, `optional`): whether to use crf, default `True`.
        **kwargs: other arguments
    """

    def __init__(
        self,
        id_to_label: Dict[int, str],
        embedder: Union[Embedder, ConfigDict],
        encoder: Optional[Union[Encoder, ConfigDict]] = None,
        dropout: float = 0.0,
        word_dropout: bool = False,
        use_crf: Optional[bool] = True,
        multiview: Optional[bool] = False,
        temperature: Optional[float] = 1.0,
        mv_loss_type: Optional[str] = 'kl',
        mv_interpolation: Optional[float] = 0.5,
        partial: Optional[bool] = False,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.id_to_label = id_to_label
        self.num_labels = len(id_to_label)

        if isinstance(embedder, Embedder):
            self.embedder = embedder
        else:
            self.embedder = Embedder.from_config(embedder)
        hidden_size = self.embedder.get_output_dim()

        if encoder is None:
            self.encoder = None
        else:
            if isinstance(encoder, Encoder):
                self.encoder = encoder
            else:
                self.encoder = Encoder.from_config(encoder)
            assert hidden_size == self.encoder.get_input_dim()
            hidden_size = self.encoder.get_output_dim()

        self.linear = nn.Linear(hidden_size, self.num_labels)

        self.use_dropout = dropout > 0.0
        if self.use_dropout:
            if word_dropout:
                self.dropout = WordDropout(dropout)
            else:
                self.dropout = nn.Dropout(dropout)

        self.use_crf = use_crf
        if use_crf:
            if partial:
                self.crf = PartialCRF(self.num_labels, batch_first=True)
            else:
                self.crf = CRF(self.num_labels, batch_first=True)
        else:
            self.loss_fn = nn.CrossEntropyLoss(reduction='mean', ignore_index=PAD_LABEL_ID)

        self.multiview = multiview
        self.mv_loss_type = mv_loss_type
        self.temperature = temperature
        self.mv_interpolation = mv_interpolation

    def forward(
        self,
        tokens: Dict[str, Any],
        label_ids: Optional[torch.LongTensor] = None,
        meta: Optional[Dict[str, Any]] = None,
        origin_tokens: Optional[Dict[str, Any]] = None,
        origin_mask: Optional[torch.Tensor] = None,
    ) -> Dict[str, Any]:
        """
        TODO docstring
        """
        logits = self._forward(tokens)

        crf_mask = get_tokens_mask(tokens, logits.size(1)) if origin_mask is None else origin_mask

        if self.training and label_ids is not None:
            loss = self._calculate_loss(logits, label_ids, crf_mask)

            if self.multiview and origin_tokens is not None:  # for multiview training
                origin_view_logits = self._forward(origin_tokens)
                origin_size = origin_view_logits.size(1 if origin_view_logits.dim() == 3 else 0)
                origin_mask_short = crf_mask[..., :origin_size]
                origin_label_ids = label_ids[..., :origin_size]

                origin_view_loss = self._calculate_loss(
                    origin_view_logits, origin_label_ids, origin_mask_short
                )
                logits_short = logits[..., :origin_size, :]
                cl_kl_loss = self._calculate_cl_loss(
                    logits_short, origin_view_logits, origin_mask_short, T=self.temperature
                )
                loss = (
                    self.mv_interpolation * (loss + origin_view_loss)
                    + (1 - self.mv_interpolation) * cl_kl_loss
                )

            outputs = {'logits': logits, 'loss': loss}
        else:
            predicts = self.decode(logits, crf_mask)
            outputs = {'logits': logits, 'predicts': predicts}

        return outputs

    def _forward(self, tokens: Dict[str, Any]) -> torch.Tensor:
        x = self.embedder(**tokens)

        if self.use_dropout:
            x = self.dropout(x)

        if self.encoder is not None:
            mask = get_tokens_mask(tokens, x.size(1))
            x = self.encoder(x, mask)

            if self.use_dropout:
                x = self.dropout(x)

        logits = self.linear(x)
        return logits

    def _calculate_loss(
        self, logits: torch.Tensor, targets: torch.Tensor, mask: torch.Tensor
    ) -> torch.Tensor:
        if self.use_crf:
            targets = targets * mask
            loss = -self.crf(logits, targets, reduction='mean', mask=mask)
        else:
            loss = self.loss_fn(logits.transpose(1, 2), targets)
        return loss

    def _calculate_cl_loss(self, ext_view_logits, origin_view_logits, mask, T=1.0):
        if self.multiview:
            batch_size, max_seq_len, num_classes = ext_view_logits.shape
            ext_view_logits = ext_view_logits.detach()
            if self.mv_loss_type == 'kl':
                _loss = (
                    F.kl_div(
                        F.log_softmax(origin_view_logits / T, dim=-1),
                        F.softmax(ext_view_logits / T, dim=-1),
                        reduction='none',
                    )
                    * mask.unsqueeze(-1)
                    * T
                    * T
                )
            elif self.mv_loss_type == 'crf_kl':
                if self.use_crf:
                    origin_view_log_posterior = self.crf.compute_posterior(origin_view_logits, mask)
                    ext_view_log_posterior = self.crf.compute_posterior(ext_view_logits, mask)
                    _loss = (
                        F.kl_div(
                            F.log_softmax(origin_view_log_posterior / T, dim=-1),
                            F.softmax(ext_view_log_posterior / T, dim=-1),
                            reduction='none',
                        )
                        * mask.unsqueeze(-1)
                        * T
                        * T
                    )
                else:
                    raise NotImplementedError
            else:
                raise NotImplementedError
            loss = _loss.sum() / batch_size
        else:
            loss = 0.0
        return loss

    def decode(  # noqa: D102
        self, logits: torch.Tensor, mask: torch.Tensor
    ) -> Union[List, torch.LongTensor]:
        if self.use_crf:
            predicts = self.crf.decode(logits, mask=mask).squeeze(0)
        else:
            predicts = logits.argmax(-1)
        return predicts



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Dict

from modelscope.metrics.base import Metric
from modelscope.metrics.builder import METRICS, MetricKeys
from modelscope.utils.tensor_utils import torch_nested_detach, torch_nested_numpify
from seqeval.metrics import accuracy_score, classification_report

from adaseq.data.constant import PAD_LABEL_ID
from adaseq.metainfo import Metrics


class SetScore:
    """evaluate macro and micro set p/r/f1 scores"""

    def __init__(self):
        self.n_sample = 0
        self.pred = []  # list of list
        self.true = []  # list of list

    def reset(self):  # noqa: D102
        self.n_sample = 0
        self.pred = []  # list of list
        self.true = []  # list of list

    def update(self, batch_gold_entities, batch_pred_entities):  # noqa: D102
        self.n_sample += len(batch_gold_entities)
        self.pred.extend(batch_pred_entities)
        self.true.extend(batch_gold_entities)

    def f1(self, precision, recall):  # noqa: D102
        f1 = 0.0 if recall + precision == 0 else (2 * precision * recall) / (precision + recall)
        return f1

    def result(self):  # noqa: D102
        assert len(self.pred) == len(self.true)
        M = len(self.pred)
        strict_acc = 0
        num_pred_labels = 0
        num_true_labels = 0
        num_correct_labels = 0
        total_ma_p = 0
        total_ma_r = 0
        total_ma_f1 = 0
        count = 0
        for i in range(M):
            p = set(self.pred[i])
            t = set(self.true[i])
            count += 1

            if p == t:
                strict_acc += 1

            l_p, l_t, l_intersect = len(p), len(t), len(p.intersection(t))
            num_pred_labels += l_p
            num_true_labels += l_t
            num_correct_labels += l_intersect

            if l_p == 0 or l_t == 0:
                ma_p = 0
                ma_r = 0
                ma_f1 = 0
            else:
                ma_p = l_intersect / l_p
                ma_r = l_intersect / l_t
                ma_f1 = self.f1(ma_p, ma_r)

            total_ma_p += ma_p
            total_ma_r += ma_r
            total_ma_f1 += ma_f1

        if num_pred_labels == 0 or num_true_labels == 0:
            micro_p = 0
            micro_r = 0
            micro_f1 = 0
        else:
            micro_p = num_correct_labels / num_pred_labels
            micro_r = num_correct_labels / num_true_labels
            micro_f1 = self.f1(micro_p, micro_r)

        strict_acc /= count
        macro_p = total_ma_p / count
        macro_r = total_ma_r / count
        macro_f1 = self.f1(macro_p, macro_r)
        avg_true_label = num_true_labels / M
        avg_pred_label = num_pred_labels / M

        return {
            'strict_acc': strict_acc,
            'micro_p': micro_p,
            'micro_r': micro_r,
            'micro_f1': micro_f1,
            'macro_p': macro_p,
            'macro_r': macro_r,
            'macro_f1': macro_f1,
            'avg_true_label': avg_true_label,
            'avg_pred_label': avg_pred_label,
        }


@METRICS.register_module(module_name=Metrics.pretraining_metric)
class PretrainingMetric(Metric):
    """Evaluate metrics for typing tasks and identification tasks"""

    def __init__(self, return_class_level_metric=False, mode=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.typing_cls_scorer = SetScore()
        self.return_class_level_metric = return_class_level_metric
        self.mode = mode
        self.ident_preds = []
        self.ident_golds = []
        self.prompt_preds = []
        self.prompt_golds = []

    def add(self, outputs: Dict, inputs: Dict):  # noqa: D102
        ident_pred_results, typing_predicts, prompt_pred_results = outputs['predicts']
        typing_pred_results = list()
        typing_ground_truths = list()
        for i, meta in enumerate(inputs['meta']):
            for j, s in enumerate(meta['spans']):
                typing_pred_results.append(typing_predicts[i][j])
                typing_ground_truths.append(set(s['type']))
        self.typing_cls_scorer.update(typing_ground_truths, typing_pred_results)

        ident_ground_truths = inputs['ident_ids']
        self.ident_preds.extend(
            torch_nested_numpify(torch_nested_detach(ident_pred_results)).tolist()
        )
        self.ident_golds.extend(
            torch_nested_numpify(torch_nested_detach(ident_ground_truths)).tolist()
        )
        prompt_ground_truths = inputs['prompt_target_label_ids']
        self.prompt_preds.extend(
            torch_nested_numpify(torch_nested_detach(prompt_pred_results)).tolist()
        )
        self.prompt_golds.extend(
            torch_nested_numpify(torch_nested_detach(prompt_ground_truths)).tolist()
        )

    def evaluate(self):  # noqa: D102

        # for detection
        id2label = self.trainer.train_preprocessor.ident_id_to_label

        pred_labels = [
            [id2label[p] for p, g in zip(ident_pred, ident_gold) if g != PAD_LABEL_ID]
            for ident_pred, ident_gold in zip(self.ident_preds, self.ident_golds)
        ]
        gold_labels = [
            [id2label[g] for g in ident_gold if g != PAD_LABEL_ID]
            for ident_gold in self.ident_golds
        ]

        ident_report = classification_report(
            gold_labels, pred_labels, output_dict=True, mode=self.mode
        )

        ident_report.pop('macro avg')
        ident_report.pop('weighted avg')
        ident_overall_score = ident_report.pop('micro avg')

        scores = {}
        scores['ident_precision'] = ident_overall_score['precision']
        scores['ident_recall'] = ident_overall_score['recall']
        scores['ident_F1'] = ident_overall_score['f1-score']
        scores['ident_accuracy'] = accuracy_score(y_true=gold_labels, y_pred=pred_labels)
        if self.return_class_level_metric:
            for tp, tp_score in ident_report.items():
                scores[tp] = {}
                scores[tp][MetricKeys.PRECISION] = round(tp_score['precision'], 4)
                scores[tp][MetricKeys.RECALL] = round(tp_score['recall'], 4)
                scores[tp][MetricKeys.F1] = round(tp_score['f1-score'], 4)
                scores[tp]['support'] = tp_score['support']
        # for typing
        typing_score_detail = self.typing_cls_scorer.result()

        scores['typing_macro_f1'] = typing_score_detail['macro_f1']
        scores['typing_macro_precision'] = typing_score_detail['macro_p']
        scores['typing_macro_recall'] = typing_score_detail['macro_r']

        scores['typing_micro_f1'] = typing_score_detail['micro_f1']
        scores['typing_micro_precision'] = typing_score_detail['micro_p']
        scores['typing_micro_recall'] = typing_score_detail['micro_r']

        scores['typing_strict_acc'] = typing_score_detail['strict_acc']
        scores['avg_true_label of typing'] = typing_score_detail['avg_true_label']
        scores['avg_pred_label of typing'] = typing_score_detail['avg_pred_label']

        scores[MetricKeys.PRECISION] = typing_score_detail['macro_p']
        scores[MetricKeys.RECALL] = typing_score_detail['macro_r']
        scores[MetricKeys.F1] = typing_score_detail['macro_f1']
        scores[MetricKeys.ACCURACY] = typing_score_detail['strict_acc']
        # prompt ner
        pred_labels = [
            [id2label[p] for p, g in zip(prompt_pred, prompt_gold) if g != PAD_LABEL_ID]
            for prompt_pred, prompt_gold in zip(self.prompt_preds, self.prompt_golds)
        ]
        gold_labels = [
            [id2label[g] for g in prompt_gold if g != PAD_LABEL_ID]
            for prompt_gold in self.prompt_golds
        ]
        prompt_report = classification_report(
            gold_labels, pred_labels, output_dict=True, mode=self.mode
        )

        prompt_report.pop('macro avg')
        prompt_report.pop('weighted avg')
        prompt_overall_score = prompt_report.pop('micro avg')

        scores['prompt_precision'] = prompt_overall_score['precision']
        scores['prompt_recall'] = prompt_overall_score['recall']
        scores['prompt_F1'] = prompt_overall_score['f1-score']
        scores['prompt_accuracy'] = accuracy_score(y_true=gold_labels, y_pred=pred_labels)

        return scores



# Copyright (c) Alibaba, Inc. and its affiliates.
import os


def is_empty_dir(path):
    """Check if a directory is empty"""
    return len(list(filter(lambda x: not x.startswith('.nfs'), os.listdir(path)))) == 0



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Any, Dict, Optional, Union

import torch
import torch.nn as nn
from modelscope.models.builder import MODELS

from adaseq.data.tokenizer import build_tokenizer
from adaseq.metainfo import Models, Tasks
from adaseq.models.base import Model
from adaseq.models.utils import bio2span
from adaseq.modules.decoders import CRF
from adaseq.modules.embedders import Embedder, TransformerEmbedder
from adaseq.modules.encoders import SpanEncoder
from adaseq.modules.util import get_tokens_mask


@MODELS.register_module(Tasks.entity_typing, module_name=Models.pretraining_model)
@MODELS.register_module(Tasks.named_entity_recognition, module_name=Models.pretraining_model)
class PretrainingModel(Model):
    """Pretraining model

    This model is used for pretraining with identification and typing.

    Args:
        num_labels (int): number of labels
        embedder (Union[embedder, str], `optional`): embedder used in the model.
            It can be an `Embedder` instance or an embedder config file or an embedder config dict.
        span_encoder_method (str): concat
        word_dropout (float, `optional`): word dropout rate, default `0.0`.
        ident_loss_function (str, `optional'): loss function, default 'PULearning',
        typing_loss_function (str, `optional'): loss function, default 'BCE',
        class_threshold (float, `optional`): classification threshold default. '0.5'
        use_biaffine (bool, `optional`): whether to use biaffine for span representation, default `False`.
        **kwargs: other arguments
    """

    def __init__(
        self,
        typing_id_to_label: Dict[int, str],
        ident_id_to_label: Dict[int, str],
        embedder: Union[Embedder, str],
        span_encoder_method: str = 'concat',
        word_dropout: Optional[float] = 0.0,
        class_threshold: float = 0.5,
        use_biaffine: bool = False,
        **kwargs
    ):
        super(PretrainingModel, self).__init__(**kwargs)
        self.ident_num_labels = len(ident_id_to_label)
        self.typing_num_labels = len(typing_id_to_label)
        self.ident_id_to_label = {int(k): v for k, v in ident_id_to_label.items()}
        self.typing_id_to_label = {int(k): v for k, v in typing_id_to_label.items()}
        if isinstance(embedder, Embedder):
            self.embedder = embedder
        else:
            self.embedder = Embedder.from_config(embedder)
            # WARNING: hard code
            self.tokenizer = build_tokenizer('bert-base-chinese')
        assert isinstance(self.embedder, TransformerEmbedder)

        self.use_dropout = word_dropout > 0.0
        if self.use_dropout:
            self.dropout = nn.Dropout(word_dropout)

        self.class_threshold = class_threshold

        self.use_biaffine = use_biaffine
        self.negative_sampling_ratio = kwargs.get('negative_sampling_ratio', 0.5)
        self.typing_penalize = kwargs.get('typing_penalize', 1.0)

        self.split_space_ident = nn.Linear(
            self.embedder.get_output_dim(), self.embedder.get_output_dim()
        )
        self.ident_linear = nn.Linear(self.embedder.get_output_dim(), self.ident_num_labels)

        self.span_encoder = SpanEncoder(
            self.embedder.get_output_dim(),
            span_encoder_method,
            use_biaffine=self.use_biaffine,
            **kwargs
        )
        self.split_space_typing = nn.Linear(
            self.embedder.get_output_dim(), self.embedder.get_output_dim()
        )
        self.typing_linear = nn.Linear(self.span_encoder.output_dim, self.typing_num_labels)
        self.split_space_prompt = nn.Linear(
            self.embedder.get_output_dim(), self.embedder.get_output_dim()
        )
        self.prompt_linear = nn.Linear(self.embedder.get_output_dim(), self.ident_num_labels)
        self.prompt_crf = CRF(self.ident_num_labels, batch_first=True)
        self.load_model_ckpt()

    def __forward_embedder(
        self, tokens: Dict[str, Any], mlm_input_ids: Optional[torch.LongTensor] = None
    ) -> torch.Tensor:
        if self.use_mlm_loss:
            assert mlm_input_ids is not None, 'Must have mlm_input_ids!!!'
            subtoken_embed = self.embedder.encode(mlm_input_ids, tokens['attention_mask'])
        else:
            subtoken_embed = self.embedder.encode(tokens['input_ids'], tokens['attention_mask'])

        if self.training and self.use_dropout:
            subtoken_embed = self.dropout(subtoken_embed)
        return subtoken_embed

    def __forward_ident(self, token_embed: torch.Tensor) -> torch.Tensor:
        ident_embed = self.split_space_ident(token_embed[:, 1:-1, :])
        ident_logits = self.ident_linear(ident_embed)
        return ident_logits

    def __forward_typing(
        self, token_embed: torch.Tensor, mention_boundary: torch.LongTensor
    ) -> torch.Tensor:
        typing_embed = self.split_space_typing(token_embed[:, 1:-1, :])
        # B*M x K
        span_reprs = self.span_encoder(typing_embed, mention_boundary)
        typing_logits = self.typing_linear(span_reprs)
        return typing_logits

    def __forward_prompt_ner(
        self, prompt_input_ids: torch.Tensor, offsets: torch.Tensor, prompt_input_mask: torch.Tensor
    ) -> torch.Tensor:
        subtoken_embed = self.embedder.encode(prompt_input_ids, prompt_input_mask)
        token_embed = self.embedder.reconstruct(subtoken_embed, offsets)
        ident_embed = self.split_space_prompt(token_embed[:, 1:-1, :])
        ident_logits = self.prompt_linear(ident_embed)
        return ident_logits

    def forward(  # noqa: D102
        self,
        tokens: Dict[str, Any],
        prompt_input_ids: Optional[torch.Tensor] = None,
        prompt_input_mask: Optional[torch.Tensor] = None,
        meta: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:  # TODO docstring
        if not self.training and prompt_input_ids is None:
            predicts = self._forward_two_stage_ner(tokens)
            outputs = {'predicts': predicts}
            return outputs
        elif not self.training and prompt_input_ids is not None:
            predicts = self._forward_zero_shot_ner(tokens, prompt_input_ids, prompt_input_mask)
            outputs = {'predicts': predicts}
            return outputs

    def _forward_two_stage_ner(  # noqa: D102
        self,
        tokens: Dict[str, Any],
        meta: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:  # TODO docstring
        # B x M -> B*M x 1
        token_embed = self.embedder.encode(tokens['input_ids'], tokens['attention_mask'])
        ident_logits = self.__forward_ident(token_embed)
        ident_predicts = ident_logits.argmax(-1)  # B*M
        pred_mentions, pred_mask = bio2span(
            ident_predicts, ident_logits.device, self.ident_id_to_label
        )

        mention_boundary = pred_mentions.transpose(-1, -2)

        typing_logits = self.__forward_typing(token_embed, mention_boundary)
        typing_predicts = typing_logits.argmax(-1)
        typing_predicts = typing_predicts.detach().cpu().numpy()

        batch_mentions = []
        mention_list = pred_mentions.tolist()
        offset_unit = len(mention_list[0])
        for b in range(len(mention_list)):
            mentions = []
            for i, mention in enumerate(mention_list[b]):
                if mention[0] == -1:
                    try:
                        assert mention[1] == -1, breakpoint()
                    except IndexError:
                        breakpoint()
                    continue
                mentions.append(
                    (
                        mention[0],
                        mention[1] + 1,
                        self.typing_id_to_label[typing_predicts[i + b * offset_unit]],
                    )
                )
            batch_mentions.append(mentions)
        return batch_mentions

    def _forward_zero_shot_ner(  # noqa: D102
        self,
        tokens: Dict[str, Any],
        prompt_input_ids: torch.Tensor,
        prompt_input_mask: torch.Tensor,
        meta: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:  # TODO docstring
        # B x M -> B*M x 1
        prompt_logits = self.__forward_prompt_ner(
            prompt_input_ids, tokens['offsets'], prompt_input_mask
        )
        label_mask = get_tokens_mask(tokens, prompt_logits.size(1))
        prompt_predicts = self.prompt_crf.decode(prompt_logits, mask=label_mask).squeeze(0)
        pred_mentions, _ = bio2span(prompt_predicts, prompt_logits.device, self.ident_id_to_label)

        batch_mentions = []
        mention_list = pred_mentions.tolist()
        for b in range(len(mention_list)):
            mentions = []
            for i, mention in enumerate(mention_list[b]):
                if mention[0] == -1:
                    try:
                        assert mention[1] == -1, breakpoint()
                    except IndexError:
                        breakpoint()
                    continue
                mentions.append((mention[0], mention[1] + 1, 'SPAN'))
            batch_mentions.append(mentions)
        return batch_mentions



# Copyright (c) Alibaba, Inc. and its affiliates.
# Copyright (c) AI2 AllenNLP. Licensed under the Apache License, Version 2.0.

from typing import Optional

import torch
from torch import nn
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

from adaseq.metainfo import Encoders
from adaseq.utils.checks import ConfigurationError

from .base import ENCODERS, Encoder


class _PytorchRnnWrapper(Encoder):
    """
    Pytorch's RNNs have two outputs: the hidden state for every time step, and the hidden state at
    the last time step for every layer.  We just want the first one as a single output.  This
    wrapper pulls out that output, and adds a `get_output_dim` method, which is useful if you
    want to, e.g., define a linear + softmax layer on top of this to get some distribution over a
    set of labels.  The linear layer needs to know its input dimension before it is called, and you
    can get that from `get_output_dim`.

    Note that we *require* you to pass a binary mask of shape (batch_size, sequence_length)
    when you call this module, to avoid subtle bugs around masking.  If you already have a
    `PackedSequence` you can pass `None` as the second parameter.
    """

    def __init__(self, module: nn.RNNBase) -> None:
        super().__init__()
        self._module = module
        if not self._module.batch_first:
            raise ConfigurationError('Our encoder semantics assumes batch is always first!')

    def get_input_dim(self) -> int:
        return self._module.input_size

    def get_output_dim(self) -> int:
        return self._module.hidden_size * (1 + int(self._module.bidirectional))

    def forward(
        self,
        inputs: torch.Tensor,
        mask: Optional[torch.BoolTensor] = None,
    ) -> torch.Tensor:
        """
        inputs: `torch.Tensor`
            Shape: [batch_size, max_length, input_dim].
        mask: `torch.BoolTensor`
            Shape: [batch_size, max_length].

        # Returns

        `torch.Tensor`
            Shape: `[batch_size, max_length, output_dim]`.
        """
        # If you already have a `PackedSequence`, you can pass `None` mask.
        if mask is None:
            return self._module(inputs)[0]

        sequence_lengths = mask.sum(-1)

        # First count how many sequences are empty.
        num_empty = torch.sum(~mask[:, 0]).int().item()
        if num_empty > 0:
            raise RuntimeError(f'Got {num_empty} empty sequences, please check your data.')

        # Now create a PackedSequence.
        packed_sequence_input = pack_padded_sequence(
            inputs, sequence_lengths.cpu(), batch_first=True, enforce_sorted=False
        )

        # Actually call the module on the sorted PackedSequence.
        packed_sequence_output, _ = self._module(packed_sequence_input)

        # unpack and restore the batch
        feature, _ = pad_packed_sequence(packed_sequence_output, batch_first=True)

        return feature


@ENCODERS.register_module(module_name=Encoders.gru_encoder)
class GruEncoder(_PytorchRnnWrapper):
    """
    Registered as a `Encoder` with name "gru".
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_layers: int = 1,
        bias: bool = True,
        dropout: float = 0.0,
        bidirectional: bool = False,
    ):
        module = torch.nn.GRU(
            input_size=input_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            bias=bias,
            batch_first=True,
            dropout=dropout,
            bidirectional=bidirectional,
        )
        super().__init__(module=module)


@ENCODERS.register_module(module_name=Encoders.lstm_encoder)
class LstmEncoder(_PytorchRnnWrapper):
    """
    Registered as a `Encoder` with name "lstm".
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_layers: int = 1,
        bias: bool = True,
        dropout: float = 0.0,
        bidirectional: bool = False,
    ):
        module = torch.nn.LSTM(
            input_size=input_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            bias=bias,
            batch_first=True,
            dropout=dropout,
            bidirectional=bidirectional,
        )
        super().__init__(module=module)


@ENCODERS.register_module(module_name=Encoders.rnn_encoder)
class RnnEncoder(_PytorchRnnWrapper):
    """
    Registered as a `Encoder` with name "rnn".
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_layers: int = 1,
        nonlinearity: str = 'tanh',
        bias: bool = True,
        dropout: float = 0.0,
        bidirectional: bool = False,
    ):
        module = torch.nn.RNN(
            input_size=input_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            nonlinearity=nonlinearity,
            bias=bias,
            batch_first=True,
            dropout=dropout,
            bidirectional=bidirectional,
        )
        super().__init__(module=module)



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Any, Dict

import numpy as np

from adaseq.metainfo import DataCollators

from .base import DATA_COLLATORS, DataCollatorWithPadding


@DATA_COLLATORS.register_module(module_name=DataCollators.span_extraction_data_collator)
class SpanExtractionDataCollatorWithPadding(DataCollatorWithPadding):
    """Padding method for span extraction dataset."""

    def padding(self, batch: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        """
        Padding a batch. In addition to the fields padded by base class
        `DataCollatorWithPadding`, label_matrix is padded here.
        """
        field = 'span_labels'
        max_length = max(len(i[0]) for i in batch[field])
        for i in range(len(batch[field])):
            difference = max_length - len(batch[field][i][0])
            if difference > 0:
                padded_labels = np.zeros((max_length, max_length), dtype=int)
                padded_labels[: batch[field][i].shape[0], : batch[field][i].shape[1]] = batch[
                    field
                ][i].astype(int)
                batch[field][i] = padded_labels.tolist()
            else:
                batch[field][i] = batch[field][i].astype(int).tolist()

        return batch



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Optional

import torch

from adaseq.data.constant import PARTIAL_LABEL_ID
from adaseq.metainfo import Decoders

from .base import DECODERS
from .crf import CRF


@DECODERS.register_module(module_name=Decoders.partial_crf)
class PartialCRF(CRF):
    """Partial/Fuzzy Conditional random field."""

    def __init__(self, num_tags: int, batch_first: bool = False) -> None:
        super().__init__(num_tags, batch_first)
        self.UNLABELED_INDEX = PARTIAL_LABEL_ID
        self.IMPOSSIBLE_SCORE = -100

    def forward(
        self,
        emissions: torch.Tensor,
        tags: torch.LongTensor,
        mask: Optional[torch.ByteTensor] = None,
        reduction: str = 'mean',
    ) -> torch.Tensor:
        """Compute the conditional log likelihood of a sequence of tags given emission scores.
        Args:
            emissions (`~torch.Tensor`): Emission score tensor of size
                ``(seq_length, batch_size, num_tags)`` if ``batch_first`` is ``False``,
                ``(batch_size, seq_length, num_tags)`` otherwise.
            tags (`~torch.LongTensor`): Sequence of tags tensor of size
                ``(seq_length, batch_size)`` if ``batch_first`` is ``False``,
                ``(batch_size, seq_length)`` otherwise.
            mask (`~torch.ByteTensor`): Mask tensor of size ``(seq_length, batch_size)``
                if ``batch_first`` is ``False``, ``(batch_size, seq_length)`` otherwise.
            reduction: Specifies  the reduction to apply to the output:
                ``none|sum|mean|token_mean``. ``none``: no reduction will be applied.
                ``sum``: the output will be summed over batches. ``mean``: the output will be
                averaged over batches. ``token_mean``: the output will be averaged over tokens.
        Returns:
            `~torch.Tensor`: The log likelihood. This will have size ``(batch_size,)`` if
            reduction is ``none``, ``()`` otherwise.
        """
        if reduction not in ('none', 'sum', 'mean', 'token_mean'):
            raise ValueError(f'invalid reduction: {reduction}')
        if mask is None:
            mask = torch.ones_like(tags, dtype=torch.uint8, device=tags.device)
        if mask.dtype != torch.uint8:
            mask = mask.byte()
        self._validate(emissions, tags=tags, mask=mask)

        # partial mask
        possible_tags = self._create_possible_tag_masks(self.num_tags, tags)

        if self.batch_first:
            emissions = emissions.transpose(0, 1)
            tags = tags.transpose(0, 1)
            mask = mask.transpose(0, 1)
            possible_tags = possible_tags.transpose(0, 1)

        # shape: (batch_size,)
        numerator = self._compute_score(emissions, tags, mask, possible_tags)
        # shape: (batch_size,)
        denominator = self._compute_normalizer(emissions, mask)
        # shape: (batch_size,)
        llh = numerator - denominator

        if reduction == 'none':
            return llh
        if reduction == 'sum':
            return llh.sum()
        if reduction == 'mean':
            return llh.mean()
        return llh.sum() / mask.float().sum()

    def _create_possible_tag_masks(self, num_tags: int, tags: torch.Tensor) -> torch.Tensor:
        # process no annotation idx
        tags = tags.clone()
        no_annotation_idx = tags == self.UNLABELED_INDEX
        tags[no_annotation_idx] = 0

        # create tag masks
        masks = torch.zeros(
            tags.size(0), tags.size(1), num_tags, dtype=torch.uint8, device=tags.device
        )
        masks.scatter_(2, tags.unsqueeze(2), 1)
        masks[no_annotation_idx] = 1  # enumerate all tags if no annotation

        return masks

    def _compute_score(
        self,
        emissions: torch.Tensor,
        tags: torch.LongTensor,
        mask: torch.ByteTensor,
        possible_tags: torch.ByteTensor,
    ) -> torch.Tensor:
        """
        Parameters:
            emissions: (seq_length, batch_size, num_tags)
            tags: (seq_length, batch_size)
            mask: (seq_length, batch_size)
            possible_tags: (seq_length, batch_size, num_tags)
        Returns:
            scores: (batch_size)
        """

        sequence_length, batch_size, num_tags = emissions.data.shape

        mask = mask.float()
        possible_tags = possible_tags.float()

        # Start transition score and first emission
        first_possible_tag = possible_tags[0]

        alpha = self.start_transitions + emissions[0]  # (batch_size, num_tags)
        alpha[(first_possible_tag == 0)] = self.IMPOSSIBLE_SCORE

        for i in range(1, sequence_length):
            current_possible_tags = possible_tags[i - 1]
            next_possible_tags = possible_tags[i]  # (batch_size, num_tags)

            # Emissions scores
            emissions_score = emissions[i].clone()
            emissions_score[(next_possible_tags == 0)] = self.IMPOSSIBLE_SCORE
            emissions_score = emissions_score.view(batch_size, 1, num_tags)

            # Transition scores
            transition_scores = (
                self.transitions.view(1, num_tags, num_tags)
                .expand(batch_size, num_tags, num_tags)
                .clone()
            )
            transition_scores[(current_possible_tags == 0)] = self.IMPOSSIBLE_SCORE
            transition_scores.transpose(1, 2)[(next_possible_tags == 0)] = self.IMPOSSIBLE_SCORE

            # Broadcast alpha
            broadcast_alpha = alpha.view(batch_size, num_tags, 1)

            # Add all scores
            inner = (
                broadcast_alpha + emissions_score + transition_scores
            )  # (batch_size, num_tags, num_tags)
            alpha = torch.logsumexp(inner, 1) * mask[i].view(batch_size, 1) + alpha * (
                1 - mask[i]
            ).view(batch_size, 1)

        # Add end transition score
        last_tag_indexes = mask.sum(0).long() - 1
        unnamed = (
            last_tag_indexes
            + torch.arange(batch_size, device=possible_tags.device) * sequence_length
        )
        end_transitions = (
            self.end_transitions.expand(batch_size, num_tags)
            * possible_tags.transpose(0, 1).view(sequence_length * batch_size, num_tags)[unnamed]
        )
        end_transitions[(end_transitions == 0)] = self.IMPOSSIBLE_SCORE
        stops = alpha + end_transitions

        return torch.logsumexp(stops, 1)  # (batch_size,)



# Copyright (c) Alibaba, Inc. and its affiliates.
import os.path as osp
from typing import Dict, Optional, Tuple

import numpy as np
import torch
import torch.nn as nn
from huggingface_hub import snapshot_download

from adaseq.metainfo import Embedders

from .base import EMBEDDERS, Embedder


@EMBEDDERS.register_module(module_name=Embedders.embedding)
class Embedding(Embedder):
    """A simple lookup table for word embedding.

    This module implements a simple lookup table that stores embeddings of a fixed dictionary and size.
    It is often used to store word embeddings and retrieve them using indices.

    It can be initialized in 2 ways:
        1) with a vocab file and a pretrained embedding file
        2) with a vocab file and embedding_dim

    The input to the module is a list of indices, and the output is the corresponding word embeddings.

    Args:
        model_name_or_path (str): Word embedding hf_repo_id or local_path.
        vocab_file (str): Vocabulary file name. Each line contains a word. Default: vocab.txt
        embedding_file (str): Embedding file name. Each line contains several rows separated by white space.
            The first row should be a word. The rest rows are floats representing the word's embedding vector.
        embedding_dim (int): Embedding dimension. Required when embedding_file is not given.
        padding_idx (int): If specified, the entries at padding_idx do not contribute to the gradient.
        freeze (bool): If True, the tensor does not get updated in the learning process.
            This works only when loading a pretrained embedding file.
    """

    def __init__(
        self,
        model_name_or_path: str,
        vocab_file: Optional[str] = 'vocab.txt',
        embedding_file: Optional[str] = 'embedding.txt',
        embedding_dim: int = 0,
        padding_idx: Optional[int] = None,
        freeze: bool = False,
    ) -> None:
        super().__init__()

        if not osp.isdir(model_name_or_path):
            local_path = snapshot_download(model_name_or_path)
        else:
            local_path = model_name_or_path

        vocab_path = osp.join(local_path, vocab_file)
        assert osp.exists(vocab_path)
        self._word2id, self._id2word = self._load_vocab(vocab_path)
        self._vocab_size = len(self._word2id)

        embedding_path = osp.join(local_path, embedding_file)
        if osp.exists(embedding_path):
            self._embedding_dim = self._infer_embedding_dim(embedding_path)
            weight = np.zeros((self._vocab_size, self._embedding_dim))
            for line in open(embedding_path):
                fields = line.strip().split(' ')
                if len(fields) == 2:
                    continue
                assert len(fields) - 1 == self._embedding_dim
                if fields[0] in self._word2id:
                    idx = self._word2id[fields[0]]
                    word_embed = [float(x) for x in fields[1:]]
                    weight[idx] = word_embed
            weight = torch.tensor(weight, dtype=torch.float32)
            self.embedding = nn.Embedding.from_pretrained(
                weight, padding_idx=padding_idx, freeze=freeze
            )
        else:
            assert embedding_dim > 0, 'must pass a `embedding_dim` if use random embedding'
            self._embedding_dim = embedding_dim
            self.embedding = nn.Embedding(
                self._vocab_size, self._embedding_dim, padding_idx=padding_idx
            )

    def forward(self, input_ids: torch.Tensor, **kwargs) -> torch.Tensor:  # noqa: D102
        return self.embedding(input_ids)

    def get_output_dim(self) -> int:
        """
        Get the output embedding dim.
        """
        return self._embedding_dim

    @property
    def embedding_dim(self) -> int:
        """Dimension of word embedding"""
        return self._embedding_dim

    @property
    def vocab_size(self) -> int:
        """Number of words in vocabulary"""
        return self._vocab_size

    @property
    def word2id(self) -> Dict[str, int]:
        """Mapping word to index"""
        return self._word2id

    @property
    def id2word(self) -> Dict[int, str]:
        """Mapping index to word"""
        return self._id2word

    @staticmethod
    def _load_vocab(vocab_file: str) -> Tuple[Dict[str, int], Dict[int, str]]:
        """Load from vocab file

        Args:
            vocab_file (str): vocab file to be loaded

        Returns:
            word2id (Dict[str, int]): a dict mapping word to id
            id2word (Dict[int, str]): a dict mapping id to word
        """
        idx = 0
        word2id = {}
        id2word = {}
        for line in open(vocab_file):
            word = line.strip()
            word2id[word] = idx
            id2word[idx] = word
            idx += 1
        return word2id, id2word

    @staticmethod
    def _infer_embedding_dim(embedding_file: str) -> int:
        """Infer embedding dimension from embedding file

        Args:
            embedding_file (str): local embedding file path

        Returns:
            embedding_dim (int): embedding dimension
        """
        embedding_dim = 0
        with open(embedding_file) as f:
            for line in f:
                fields = line.strip().split(' ')
                if len(fields) == 2:
                    continue
                embedding_dim = len(fields) - 1
                break
        assert embedding_dim > 0
        return embedding_dim



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Any, Dict

from adaseq.metainfo import DataCollators

from .base import DATA_COLLATORS, DataCollatorWithPadding


@DATA_COLLATORS.register_module(module_name=DataCollators.multi_label_span_typing_data_collator)
class MultiLabelSpanTypingDataCollatorWithPadding(DataCollatorWithPadding):
    """Padding method for multilabel span typing dataset."""

    def padding(self, batch: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        """
        Padding a batch. In addition to the fields padded by base class DataCollatorWithPadding,
        'mention_boundary'、'type_ids'、'mention_mask' are padded here.
        """

        max_span_count = max([len(x[0]) for x in batch['mention_boundary']])
        for i in range(len(batch['mention_boundary'])):
            difference = max_span_count - len(batch['mention_boundary'][i][0])
            if difference > 0:
                batch['mention_boundary'][i][0] = batch['mention_boundary'][i][0] + [0] * difference
                batch['mention_boundary'][i][1] = batch['mention_boundary'][i][1] + [0] * difference
                batch['type_ids'][i] = (
                    batch['type_ids'][i] + ([[0] * len(batch['type_ids'][i][0])]) * difference
                )
                batch['mention_mask'][i] = batch['mention_mask'][i] + [0] * difference
        return batch


@DATA_COLLATORS.register_module(module_name=DataCollators.multi_label_concat_typing_data_collator)
class MultiLabelConcatTypingDataCollatorWithPadding(DataCollatorWithPadding):
    """Padding method for multilabel span concat typing dataset."""

    def __init__(self, tokenizer, **kwargs):
        super().__init__(tokenizer, **kwargs)
        self.no_pad_fields.add('type_ids')



# Copyright (c) Alibaba, Inc. and its affiliates.
# trigger register mechanism
from adaseq import data, metrics, models, ms_patch, pipelines, training
from adaseq.utils.logging import prepare_global_logging
from adaseq.version import __version__

prepare_global_logging()



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Any, Dict, Optional, Union

import torch
import torch.nn as nn
from modelscope.models.builder import MODELS

from adaseq.data.constant import PAD_LABEL_ID
from adaseq.metainfo import Models, Tasks
from adaseq.models.base import Model
from adaseq.modules.decoders.crf import CRFwithConstraints
from adaseq.modules.embedders import Embedder
from adaseq.modules.encoders import SpanEncoder
from adaseq.modules.util import get_tokens_mask


@MODELS.register_module(Tasks.entity_typing, module_name=Models.twostage_ner_model)
@MODELS.register_module(Tasks.named_entity_recognition, module_name=Models.twostage_ner_model)
class TwoStageNERModel(Model):
    """Two stage ner model

    This model is used for two-stage few-shot ner model with pre-trained NER model (see pretraining_model module).

    Args:
        typing_id_to_label (dict): id2label for typing
        ident_id_to_label (dict): id2label for mention identification
        embedder (Union[Encoder, str], `optional`): embedder used in the model.
            It can be an `embedder` instance or an embedder config file or an encoder config dict.
        span_encoder_method (str): concat
        word_dropout (float, `optional`): word dropout rate, default `0.0`.
        use_biaffine (bool, `optional`): whether to use biaffine for span representation, default `False`.
        **kwargs: other arguments
    """

    def __init__(
        self,
        typing_id_to_label: Dict[int, str],
        ident_id_to_label: Dict[int, str],
        embedder: Union[Embedder, str],
        use_crf: bool = True,
        load_parts: str = None,
        model_helper_path: str = None,
        span_encoder_method: str = 'concat',
        word_dropout: Optional[float] = 0.0,
        use_biaffine: bool = False,
        **kwargs
    ):
        super(TwoStageNERModel, self).__init__()
        self.ident_num_labels = len(ident_id_to_label)
        self.typing_num_labels = len(typing_id_to_label)
        self.ident_id_to_label = ident_id_to_label
        self.typing_id_to_label = typing_id_to_label

        self.use_biaffine = use_biaffine

        if isinstance(embedder, Embedder):
            self.embedder = embedder
        else:
            self.embedder = Embedder.from_config(embedder)

        self.span_encoder = SpanEncoder(
            self.embedder.config.hidden_size,
            span_encoder_method,
            use_biaffine=self.use_biaffine,
            **kwargs
        )

        self.ident_linear = nn.Linear(self.embedder.get_output_dim(), self.ident_num_labels)
        self.typing_linear = nn.Linear(self.span_encoder.output_dim, self.typing_num_labels)

        self.split_space_ident = None
        self.split_space_typing = None
        if load_parts is not None and model_helper_path is not None:
            self.load_from_pretrained(load_parts, model_helper_path)

        self.use_dropout = word_dropout > 0.0
        if self.use_dropout:
            self.dropout = nn.Dropout(word_dropout)

        self.use_crf = use_crf

        if self.use_crf:
            ident_id2label_list = [v for k, v in ident_id_to_label.items()]
            self.crf = CRFwithConstraints(
                ident_id2label_list, batch_first=True, add_constraint=True
            )
        else:
            self.ident_loss_fn = nn.CrossEntropyLoss(reduction='mean', ignore_index=PAD_LABEL_ID)

        self.typing_loss_fn = nn.CrossEntropyLoss(reduction='none')

    def load_from_pretrained(self, loaded_parts, model_helper_path):
        """load pretrained model"""

        loaded_list = loaded_parts.split('I')

        pretrained_state_dict = torch.load(model_helper_path)

        if 'e' in loaded_list:
            self.embedder.load_state_dict(pretrained_state_dict['embedder'])

        if 'i' in loaded_list:
            self.ident_linear.load_state_dict(pretrained_state_dict['ident_linear'])

        if 'si' in loaded_list:
            self.split_space_ident = nn.Linear(
                self.embedder.config.hidden_size, self.embedder.config.hidden_size
            )
            self.split_space_ident.load_state_dict(pretrained_state_dict['split_space_ident'])

        if 'st' in loaded_list:
            self.split_space_typing = nn.Linear(
                self.embedder.config.hidden_size, self.embedder.config.hidden_size
            )
            self.split_space_typing.load_state_dict(pretrained_state_dict['split_space_typing'])

    def _forward_embedder(self, tokens: Dict[str, Any]) -> torch.Tensor:
        token_embed = self.embedder(**tokens)

        if self.use_dropout:
            token_embed = self.dropout(token_embed)

        return token_embed

    def _forward_ident(self, token_embed: torch.Tensor) -> torch.Tensor:

        # for entity detection
        ident_embed = token_embed
        if self.split_space_ident is not None:
            ident_embed = self.split_space_ident(token_embed)

        ident_logits = self.ident_linear(ident_embed)

        return ident_logits

    def _forward_typing(
        self, token_embed: torch.Tensor, mention_boundary: torch.Tensor
    ) -> torch.Tensor:

        # for entity typing
        typing_embed = token_embed
        if self.split_space_typing is not None:
            typing_embed = self.split_space_typing(token_embed)

        # B*M x K
        span_reprs = self.span_encoder(typing_embed, mention_boundary)
        typing_logits = self.typing_linear(span_reprs)

        return typing_logits

    def forward(  # noqa: D102
        self,
        tokens: Dict[str, Any],
        mention_boundary: torch.LongTensor,
        mention_mask: torch.LongTensor,
        ident_ids: Optional[torch.LongTensor] = None,
        type_ids: Optional[torch.LongTensor] = None,
        meta: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:  # TODO docstring

        token_embed = self._forward_embedder(tokens)
        ident_mask = get_tokens_mask(tokens, token_embed.size(1))

        if self.training:
            ident_logits = self._forward_ident(token_embed)
            typing_logits = self._forward_typing(token_embed, mention_boundary)
            logits = (ident_logits, typing_logits)
            loss = self._calculate_loss(logits, ident_ids, type_ids, ident_mask, mention_mask)
            outputs = {'logits': logits, 'loss': loss}
        else:
            predicts = self.decode(token_embed, ident_mask)
            outputs = {'predicts': predicts}
        return outputs

    def _calculate_loss(self, logits, ident_target, typing_target, ident_mask, typing_mask):
        """
        Calculate loss for two-stage model
        """

        ident_loss = typing_loss = 0

        ident_logits, typing_logits = logits

        if self.use_crf:
            ident_target = ident_target * ident_mask
            ident_loss = -self.crf(ident_logits, ident_target, reduction='mean', mask=ident_mask)
        else:
            ident_loss = self.ident_loss_fn(
                ident_logits.reshape(-1, self.ident_num_labels), ident_target.reshape(-1)
            )

        typing_loss = self._span_label_loss(typing_logits, typing_target, typing_mask)

        loss = ident_loss + typing_loss

        return loss

    def _span_label_loss(self, typing_logits, typing_target, typing_mask):

        num_types = typing_logits.size(-1)

        typing_target = typing_target.reshape(-1).to(torch.long)  # B*M x L
        typing_logits = typing_logits.reshape(-1, typing_logits.size(-1))
        typing_mask = typing_mask.reshape(-1).bool()

        typing_logits = typing_logits.masked_select(typing_mask.unsqueeze(-1)).reshape(
            -1, num_types
        )
        typing_target = typing_target.masked_select(typing_mask)
        typing_loss = self.typing_loss_fn(typing_logits, typing_target)
        typing_loss = typing_loss.sum() / typing_mask.sum()

        return typing_loss

    def bio2span(self, sequences, device):
        """label sequence to span format"""

        bsz = sequences.size(0)

        all_spans = [[] for _ in range(bsz)]

        for b in range(len(all_spans)):
            all_spans[b] = self._convert2span(sequences[b].tolist())

        max_num_spans = max(len(i) for i in all_spans)

        if max_num_spans == 0:
            max_num_spans = 1

        pred_mentions = [i + (max_num_spans - len(i)) * [[-1, -1]] for i in all_spans]

        pred_mentions = torch.tensor(pred_mentions).to(device)
        pred_mask = pred_mentions[..., 0] != -1

        return pred_mentions, pred_mask

    def _convert2span(self, label_list):
        span_list = []
        i = 0
        label_list = [self.ident_id_to_label[item] for item in label_list]
        while i < len(label_list):
            if label_list[i][0] == 'B' or label_list[i][0] == 'I':
                start_idx = i
                i += 1
                while i < len(label_list) and not (
                    label_list[i][0] == 'O' or label_list[i][0] == 'E'
                ):
                    i += 1
                if i < len(label_list):
                    end_idx = i if label_list[i][0] == 'E' else i - 1
                    # Looks like a good trick
                    span_list.append([start_idx, end_idx])
                    i += 1
                else:
                    span_list.append([start_idx, i - 1])

            elif label_list[i][0] == 'S' or label_list[i][0] == 'E':
                span_list.append([i, i])
                i += 1
            else:
                i += 1

        return span_list

    def decode(self, embed, ident_mask):  # noqa: D102

        # for detection

        ident_logits = self._forward_ident(embed)
        if self.use_crf:
            ident_predicts = self.crf.decode(ident_logits, mask=ident_mask).squeeze(0)
        else:
            ident_predicts = ident_logits.argmax(-1)  # B*M

        pred_mentions, pred_mask = self.bio2span(ident_predicts, ident_logits.device)

        mention_boundary = pred_mentions.transpose(-1, -2)

        typing_logits = self._forward_typing(embed, mention_boundary)
        typing_predicts = typing_logits.argmax(-1)
        typing_predicts = typing_predicts.detach().cpu().numpy()

        batch_mentions = []
        mention_list = pred_mentions.tolist()
        offset_unit = len(mention_list[0])
        for b in range(len(mention_list)):
            mentions = []
            for i, mention in enumerate(mention_list[b]):
                if mention[0] == -1:
                    try:
                        assert mention[1] == -1, breakpoint()
                    except IndexError:
                        breakpoint()
                    continue
                mentions.append(
                    (
                        mention[0],
                        mention[1] + 1,
                        self.typing_id_to_label[typing_predicts[i + b * offset_unit]],
                    )
                )
            batch_mentions.append(mentions)

        return batch_mentions



# Copyright (c) Alibaba, Inc. and its affiliates.
# Some codes are modified from AllenNLP.
# Copyright (c) AI2 AllenNLP. Licensed under the Apache License, Version 2.0.

from typing import Dict, List, Optional, Tuple

TypedSpan = Tuple[int, int, str]


class InvalidTagSequence(Exception):  # noqa: D101
    def __init__(self, tags=None):
        super().__init__()
        self.tags = tags or list()

    def __str__(self):
        return ' '.join(self.tags)


def bio_tags_to_spans(tags: List[str], classes_to_ignore: Optional[List[str]] = None) -> List[Dict]:
    """
    Given a sequence corresponding to BIO tags, extracts spans.
    Span start is inclusive (closed), end is exclusive (open).
    This function works properly when the spans are unlabeled (i.e., your labels
    are simply "B", "I", and "O").
    # Parameters
    tags : `List[str]`, required.
        The string class labels for a sequence.
    classes_to_ignore : `List[str]`, optional (default = `None`).
        A list of string class labels `excluding` the bio tag
        which should be ignored when extracting spans.
    # Returns
    spans : `List[Dict]`
        The typed, extracted spans from the sequence, in the format
        `{'start': start, 'end': end, 'type': label}`.
        Note that the label `does not` contain any BIO tag prefixes.
    """
    classes_to_ignore = classes_to_ignore or list()
    spans: List[Dict] = list()
    start = 0
    end = 0
    active_category = None
    for index, string_tag in enumerate(tags):
        # Actual BIO tag.
        bio_tag = string_tag[0]
        if bio_tag not in 'BIO':
            raise InvalidTagSequence(tags)
        category = string_tag[2:]
        if bio_tag == 'O' or category in classes_to_ignore:
            # The span has ended.
            if active_category is not None:
                spans.append(dict(start=start, end=end + 1, type=active_category))
            active_category = None
            continue
        elif bio_tag == 'B':
            # We are entering a new span; reset indices and active tag to new span.
            if active_category is not None:
                spans.append(dict(start=start, end=end + 1, type=active_category))
            active_category = category
            start = index
            end = index
        elif bio_tag == 'I' and category == active_category:
            # We're inside a span.
            end += 1
        else:
            raise InvalidTagSequence(tags)
    # Last token might have been a part of a valid span.
    if active_category is not None:
        spans.append(dict(start=start, end=end + 1, type=active_category))
    return spans



# Copyright (c) Alibaba, Inc. and its affiliates.
import os
import shutil
from typing import Optional

import yaml
from modelscope.utils.config import Config
from modelscope.utils.torch_utils import set_random_seed

from adaseq.data.data_collators.base import build_data_collator
from adaseq.data.dataset_manager import DatasetManager
from adaseq.data.preprocessors.nlp_preprocessor import build_preprocessor
from adaseq.metainfo import Trainers
from adaseq.training import build_trainer
from adaseq.utils.checks import ConfigurationError
from adaseq.utils.common_utils import create_datetime_str
from adaseq.utils.file_utils import is_empty_dir
from adaseq.utils.logging import prepare_logging


def pretrain_model(
    config_path: str,
    work_dir: Optional[str] = None,
    run_name: Optional[str] = None,
    seed: Optional[int] = None,
    force: bool = False,
    device: str = 'gpu',
    local_rank: str = '0',
    checkpoint_path: Optional[str] = None,
) -> None:
    """
    Train a model from config file.
    You can mannualy call this function in a python script for debugging.
    """
    config = Config.from_file(config_path)

    # create work_dir
    if work_dir is None:
        work_dir = os.path.join(
            config.safe_get('experiment.exp_dir', 'experiments/'),
            config.safe_get('experiment.exp_name', 'unknown/'),
            run_name or create_datetime_str(),
        )
    if os.path.exists(work_dir) and force:
        shutil.rmtree(work_dir, ignore_errors=True)
    if os.path.exists(work_dir) and not is_empty_dir(work_dir):
        raise ConfigurationError(f'`work_dir` ({work_dir}) already exists and is not empty.')
    os.makedirs(work_dir, exist_ok=True)

    world_size = int(os.environ['WORLD_SIZE']) if 'WORLD_SIZE' in os.environ else 1
    prepare_logging(work_dir, int(local_rank), world_size)

    # Get seed from the comand line args first.
    if seed is None:
        # if not given, try to get one from config file
        seed = config.safe_get('experiment.seed', 42)  # 42 by default
        if seed < 0:
            raise ConfigurationError(f'random seed must be greater than 0, got: {seed}')
    else:
        config['experiment']['seed'] = seed
    set_random_seed(seed)

    trainer = build_pretrainer_from_partial_objects(
        config,
        work_dir=work_dir,
        seed=seed,
        device=device,
        local_rank=local_rank,
    )
    trainer.train(checkpoint_path)
    trainer.test()


def build_pretrainer_from_partial_objects(config, work_dir, **kwargs):
    """
    Entrypoint of build the trainer from `config` by modelscope.
    In this method, we will build the `DatasetManager` first, then use the
    counted or loaded `labels` to build the`Preprocessor`.
    The the`Preprocessor` will provide the final `id_to_label` mapping,
    which is a required argument of all `AdaSeq` models, we update it to the
    `model` section of `config` and dump the updated `config` to the `work_dir`.

    Args:

    work_dir (`str`): required
        The created directionary to save all produced files in training.
    config (`Config`): required
        The `Config` of this trial.
    """
    # build datasets via `DatasetManager`
    dm = DatasetManager.from_config(task=config.task, **config.dataset)
    # build preprocessor with config and labels
    preprocessor = build_preprocessor(config.preprocessor, labels=dm.labels)

    if 'lr_scheduler' not in config.train:  # default constant lr.
        config.train['lr_scheduler'] = dict(type='constant')

    # Finally, get `id_to_label` for model.
    config.model.typing_id_to_label = preprocessor.typing_id_to_label
    config.model.ident_id_to_label = preprocessor.ident_id_to_label
    # Dump config to work_dir and reload.
    new_config_path = os.path.join(work_dir, 'config.yaml')
    with open(new_config_path, mode='w', encoding='utf8') as file:
        yaml.dump(config.to_dict(), file, allow_unicode=True)

    # build `DataCollator` from config and tokenizer.
    collator_config = config.data_collator
    if isinstance(collator_config, str):
        collator_config = dict(type=collator_config)
    data_collator = build_data_collator(preprocessor.tokenizer, collator_config)

    trainer = build_trainer(
        config.safe_get('train.trainer', Trainers.default_trainer),
        cfg_file=new_config_path,
        work_dir=work_dir,
        dataset_manager=dm,
        data_collator=data_collator,
        preprocessor=preprocessor,
        **kwargs,
    )
    return trainer



# Copyright (c) Alibaba, Inc. and its affiliates.
import os
from abc import ABC, abstractmethod

import datasets
from datasets import DownloadManager
from datasets.utils.file_utils import is_remote_url


class CustomDatasetBuilder(ABC, datasets.GeneratorBasedBuilder):
    """Base class for custumized dataset builder."""

    @abstractmethod
    def stub():
        """Useless stub function.

        The datasets.load_dataset import the builder class from path via import_main_class.
        According to the source code, import_main_class will return the first non-abstract class.
        So we should make this base class a abstract class by adding this stub function.
        Otherwise, this base class will be returned by import_main_class,simce it is imported
        before the definition of sub-classes.
        """

        pass

    def _split_generators(self, dl_manager: DownloadManager):
        """Specify feature dictionary generators and dataset splits.

        This function returns a list of `SplitGenerator`s defining how to generate
        data and what splits to use.

        Example:

            return[
                    datasets.SplitGenerator(
                            name=datasets.Split.TRAIN,
                            gen_kwargs={'file': 'train_data.zip'},
                    ),
                    datasets.SplitGenerator(
                            name=datasets.Split.TEST,
                            gen_kwargs={'file': 'test_data.zip'},
                    ),
            ]

        The above code will first call `_generate_examples(file='train_data.zip')`
        to write the train data, then `_generate_examples(file='test_data.zip')` to
        write the test data.

        Datasets are typically split into different subsets to be used at various
        stages of training and evaluation.

        Note that for datasets without a `VALIDATION` split, you can use a
        fraction of the `TRAIN` data for evaluation as you iterate on your model
        so as not to overfit to the `TEST` data.

        For downloads and extractions, use the given `download_manager`.
        Note that the `DownloadManager` caches downloads, so it is fine to have each
        generator attempt to download the source data.

        A good practice is to download all data in this function, and then
        distribute the relevant parts to each split with the `gen_kwargs` argument

        Args:
            dl_manager: (DownloadManager) Download manager to download the data

        Returns:
            `list<SplitGenerator>`.
        """

        if self.config.data_dir is not None:
            if is_remote_url(self.config.data_dir):
                _data_dir = dl_manager.download_and_extract(self.config.data_dir)
            elif not os.path.isdir(self.config.data_dir):
                # could be some archieve files like `DIR/FILE.zip`
                _data_dir = dl_manager.extract(self.config.data_dir)
            else:
                _data_dir = self.config.data_dir

            data_files = dict()
            _all_files = os.listdir(_data_dir)
            for split_name in ['train', 'valid', 'test']:
                _file = get_file_by_keyword(_all_files, split_name)
                if _file is None and split_name == 'valid':
                    _file = get_file_by_keyword(_all_files, 'dev')
                if _file is None:
                    continue
                data_files[split_name] = os.path.join(_data_dir, _file)

        elif self.config.data_files is not None:
            assert isinstance(self.config.data_files, dict)

            data_files = dict()
            for k, v in self.config.data_files.items():
                if isinstance(v, list):
                    v = v[0]
                if not isinstance(v, str):
                    v = v.as_posix()
                if is_remote_url(v):
                    v = dl_manager.download(v)
                data_files[k] = v
        else:
            raise ValueError('Datasets cannot be resolved!')

        return [
            datasets.SplitGenerator(
                name=split_name, gen_kwargs={'filepath': data_files[split_name]}
            )
            for split_name in data_files.keys()
        ]


def get_file_by_keyword(files, keyword):
    """Get file by keyword, such as: train/test/dev."""

    for filename in files:
        if keyword in filename:
            return filename
    return None



# Copyright (c) Alibaba, Inc. and its affiliates.
import os
from typing import Dict, Optional

from modelscope.metrics.base import Metric


class DatasetDumper(Metric):
    """class to dump model predictions"""

    def __init__(self, save_path: Optional[str] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.save_path = save_path
        self.data = []

    def add(self, outputs: Dict, inputs: Dict):
        """add predictions to cache"""
        raise NotImplementedError

    def evaluate(self):
        """creat dump file and dump predicitons."""
        if self.save_path is None:
            self.save_path = os.path.join(self.trainer.work_dir, 'pred.txt')
        save_path = self.save_path
        self.save_path = save_path + '.tmp'  # file lock
        self.dump()
        os.replace(self.save_path, save_path)
        self.save_path = save_path
        return {}

    def dump(self):
        """dump predictions"""
        raise NotImplementedError



#!/usr/bin/env python
# Copyright (c) Alibaba, Inc. and its affiliates.
import os
import sys

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.realpath(__file__))))


def run():
    """Command line main interface"""
    from adaseq.commands import main

    main(prog='adaseq')


if __name__ == '__main__':
    run()



# Copyright (c) Alibaba, Inc. and its affiliates.
import json
import logging
import os
import os.path as osp
from collections import OrderedDict

from modelscope.metainfo import Hooks
from modelscope.trainers.hooks.builder import HOOKS
from modelscope.trainers.hooks.logger.text_logger_hook import TextLoggerHook
from modelscope.utils.json_utils import EnhancedEncoder
from modelscope.utils.torch_utils import is_master

logger = logging.getLogger(__name__)


@HOOKS.register_module(module_name=Hooks.TextLoggerHook, force=True)
class AdaSeqTextLoggerHook(TextLoggerHook):
    """Logger hook in text, Output log to both console and local json file.

    Args:
        by_epoch (bool, optional): Whether EpochBasedtrainer is used.
            Default: True.
        interval (int, optional): Logging interval (every k iterations).
            It is interval of iterations even by_epoch is true. Default: 10.
        ignore_last (bool, optional): Ignore the log of last iterations in each
            epoch if less than :attr:`interval`. Default: True.
        reset_flag (bool, optional): Whether to clear the output buffer after
            logging. Default: False.
        out_dir (str): The directory to save log. If is None, use `trainer.work_dir`
        ignore_rounding_keys (`Union[str, List]`): The keys to ignore float rounding, default 'lr'
        rounding_digits (`int`): The digits of rounding, exceeding parts will be ignored.
    """

    def __init__(
        self,
        by_epoch=True,
        interval=10,
        ignore_last=True,
        reset_flag=False,
        out_dir=None,
        ignore_rounding_keys='lr',
        rounding_digits=5,
        filename: str = 'metrics',
    ) -> None:
        super().__init__(
            by_epoch=by_epoch,
            interval=interval,
            ignore_last=ignore_last,
            reset_flag=reset_flag,
            out_dir=out_dir,
            ignore_rounding_keys=ignore_rounding_keys,
            rounding_digits=rounding_digits,
        )
        self.filename = filename

    def before_run(self, trainer):  # noqa: D102
        super(TextLoggerHook, self).before_run(trainer)

        if self.out_dir is None:
            self.out_dir = trainer.work_dir

        if not osp.exists(self.out_dir) and is_master():
            os.makedirs(self.out_dir)

        self.start_iter = trainer.iter
        self.json_log_path = osp.join(self.out_dir, f'{self.filename}.json')

        logger.info('Text logs will be saved to: %s', self.json_log_path)

    def _dump_log(self, log_dict):
        # dump log in json format
        json_log = OrderedDict()
        for k, v in log_dict.items():
            json_log[k] = (
                v if k in self.ignore_rounding_keys else self._round_float(v, self.rounding_digits)
            )

        if is_master():
            with open(self.json_log_path, 'a+') as f:
                json.dump(json_log, f, cls=EnhancedEncoder, ensure_ascii=False)
                f.write('\n')



# Copyright (c) Alibaba, Inc. and its affiliates.
import argparse
from typing import Optional, Tuple

from adaseq import __version__
from adaseq.commands.test import Test
from adaseq.commands.train import Train


def parse_args(  # noqa: D103
    prog: Optional[str] = None,
) -> Tuple[argparse.ArgumentParser, argparse.Namespace]:
    parser = argparse.ArgumentParser(prog=prog)
    parser.add_argument('--version', action='version', version=f'%(prog)s {__version__}')

    subparsers = parser.add_subparsers(help='commands')
    Train.add_subparser(subparsers)
    Test.add_subparser(subparsers)

    args = parser.parse_args()

    return parser, args


def main(prog: Optional[str] = None) -> None:  # noqa: D103
    parser, args = parse_args(prog)

    if 'func' in dir(args):
        args.func(args)
    else:
        parser.print_help()



# Copyright (c) Alibaba, Inc. and its affiliates.
from collections.abc import Mapping

import torch


class DataBatch(Mapping):
    """represent a data batch, support `tensorize` function."""

    def __init__(self, batch, keep_fields=None):
        self.keep_fields = keep_fields or {}
        # modelscope/utils/data_utils.py: 17 会调这个 __init__,
        # 传入参数只有dict类型的batch, 但是value其实都是tensorize好的，
        # 所以不需要再做一次。
        if len(self.keep_fields) == 0:
            if isinstance(batch, tuple):
                batch = dict(batch)
            self.batch = batch
        else:
            self.batch = self.tensorize(batch)

    def __repr__(self):
        return str(self.batch)

    def __getitem__(self, key):
        return self.batch.get(key, None)

    def __contains__(self, key):
        return key in self.batch

    def __iter__(self):
        return iter(self.batch)

    def __len__(self):
        return len(self.batch)

    def tensorize(self, batch):
        """convert all possible fields to tensor."""

        if isinstance(batch, tuple):
            return dict(batch)
        for k in batch.keys():
            if k.endswith('tokens'):
                for sub in batch[k].keys():
                    dtype = torch.bool if sub.endswith('mask') else torch.long
                    batch[k][sub] = torch.tensor(batch[k][sub], dtype=dtype)
            elif k in self.keep_fields:
                continue
            else:
                batch[k] = torch.tensor(batch[k])
        return batch

    def to(self, device):
        """move to device"""
        for k in self.batch.keys():
            if k.endswith('tokens'):
                for sub in self.batch[k].keys():
                    self.batch[k][sub] = self.batch[k][sub].to(device)
            elif k in self.keep_fields:
                continue
            else:
                self.batch[k] = self.batch[k].to(device)



# Copyright (c) Alibaba, Inc. and its affiliates.

from .biaffine_ner_model import BiaffineNerModel
from .global_pointer_model import GlobalPointerModel
from .multilabel_typing_model import (
    MultiLabelConcatTypingModel,
    MultiLabelSpanTypingModel,
)
from .pretraining_model import PretrainingModel
from .relation_extraction_model import RelationExtractionModel
from .sequence_labeling_model import SequenceLabelingModel
from .twostage_ner_model import TwoStageNERModel



from typing import Optional

import torch
import torch.nn as nn
from torch.nn import BCELoss, BCEWithLogitsLoss


class WBCEWithLogitsLoss:
    """Weighed BCE loss, multiply the loss of positive examples with a scaler"""

    def __init__(self, pos_weight=1.0):
        self.pos_weight = pos_weight
        self.loss = BCEWithLogitsLoss(reduction='none')

    def __call__(self, y_pred, y_true):  # noqa: D102
        loss = self.loss(y_pred, y_true.float())

        if self.pos_weight != 1:
            weights_pr = torch.ones_like(y_true).float()  # B x C
            weights_pr[y_true > 0] = self.pos_weight
            loss = (loss * weights_pr).mean()
        else:
            loss = loss.mean()

        return loss


class WBCEWithLogitsLossUFET:
    """Weighed BCE loss, multiply the loss of positive examples with a scaler,
    Apply the trick in Ultra-Fine Entity Typing with Weak Supervision from a Masked Language Model"""

    def __init__(self, pos_weight=1.0):
        self.pos_weight = pos_weight
        self.loss = BCEWithLogitsLoss(reduction='none')

    def __call__(self, y_pred, y_true):  # noqa: D102
        loss = self.loss(y_pred, y_true.float())
        B, L = y_true.shape
        coarse = y_true[:, :9].sum(-1) > 0
        fine = y_true[:, 9:130].sum(-1) > 0
        ultra = y_true[:, 130:].sum(-1) > 0
        a = coarse.view(-1, 1) * torch.ones_like(y_true[:, :9])
        b = fine.view(-1, 1) * torch.ones_like(y_true[:, 9:130])
        c = ultra.view(-1, 1) * torch.ones_like(y_true[:, 130:])
        weight_cfu = torch.cat([a, b, c], -1)
        loss = loss * weight_cfu
        if self.pos_weight != 1:
            weights_pr = torch.ones_like(y_true).float()  # B x C
            weights_pr[y_true > 0] = self.pos_weight
            loss = (loss * weights_pr).mean()
        else:
            loss = loss.mean()

        return loss


class PartialBCELoss(BCELoss):
    """PartialBCELoss"""

    def __init__(
        self,
        weight=None,
        size_average=None,
        reduce=None,
        reduction='none',
        total_num_labels=0,
        negative_sampling=0.3,
    ):
        super(PartialBCELoss, self).__init__(weight, size_average, reduce, reduction)
        self.num_labels = total_num_labels
        self.negative_sampling = negative_sampling

    def forward(self, input, target):  # noqa
        return self.partial_binary_cross_entropy(input, target)

    def partial_binary_cross_entropy(self, input, target):  # noqa
        # input: B*M x L
        # target: B*M x L
        assert input.shape[0] == target.shape[0] and input.shape[1] == target.shape[1]
        if self.negative_sampling > 0:
            # breakpoint()
            negative_nums = int(self.num_labels * self.negative_sampling)
            sel_negative_ids = torch.randint(0, self.num_labels, (negative_nums, 1)).squeeze()
            sel_negatives = torch.tensor([0] * self.num_labels).to(target.device)
            sel_negatives[sel_negative_ids] = 1
            loss = -(
                target * torch.log(input)
                + (1 - target) * torch.log(1 - input) * sel_negatives.unsqueeze(0)
            )
            # loss = loss / negative_nums
        else:
            loss = -target * torch.log(input)
        # loss = loss / target.shape[1]
        if self.reduction == 'sum':
            loss = loss.sum()
        if self.reduction == 'mean':
            loss = loss.sum() / (target.shape[0])
        return loss


class PULearningLoss(nn.Module):
    """PU Learning Loss.
    https://aclanthology.org/P19-1231.pdf
    https://aclanthology.org/2020.findings-emnlp.60.pdf
    """

    def __init__(self, num_tags, label2id: dict, batch_first: bool = False) -> None:
        if num_tags <= 0:
            raise ValueError(f'invalid number of tags: {num_tags}')
        super().__init__()
        self.num_tags = num_tags
        self.label2id = label2id
        self.batch_first = batch_first
        self.new_tag = {item for item in label2id if item != 'O'}
        self.prior = {item: 0.05 for item in label2id if item != 'O'}
        self.balance = {item: 0.8 for item in label2id if item != 'O'}
        # self.new_tag = {'B-SPAN', 'I-SPAN', 'E-SPAN', 'S-SPAN'}
        # self.prior   = {'B-SPAN': 0.05, 'I-SPAN': 0.05,'E-SPAN': 0.05, 'S-SPAN': 0.05}
        # self.balance = {'B-SPAN': 0.8, 'I-SPAN': 0.8, 'E-SPAN': 0.8, 'S-SPAN': 0.8}
        self.gamma = 1.0

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(num_tags={self.num_tags})'

    def masked_sum(self, logits, mask):
        """sum with mask"""

        if mask.sum() == 0:
            return torch.sum(logits * mask)
        return torch.sum(logits * mask) / mask.sum()

    def forward(
        self,
        emissions: torch.Tensor,
        tags: torch.LongTensor,
        mask: Optional[torch.ByteTensor] = None,
        reduction: str = 'mean',
    ) -> torch.Tensor:
        """Compute the conditional log likelihood of a sequence of tags given emission scores.
        Args:
            emissions (`~torch.Tensor`): Emission score tensor of size
                ``(seq_length, batch_size, num_tags)`` if ``batch_first`` is ``False``,
                ``(batch_size, seq_length, num_tags)`` otherwise.
            tags (`~torch.LongTensor`): Sequence of tags tensor of size
                ``(seq_length, batch_size)`` if ``batch_first`` is ``False``,
                ``(batch_size, seq_length)`` otherwise.
            mask (`~torch.ByteTensor`): Mask tensor of size ``(seq_length, batch_size)``
                if ``batch_first`` is ``False``, ``(batch_size, seq_length)`` otherwise.
            reduction: Specifies  the reduction to apply to the output:
                ``none|sum|mean|token_mean``. ``none``: no reduction will be applied.
                ``sum``: the output will be summed over batches. ``mean``: the output will be
                averaged over batches. ``token_mean``: the output will be averaged over tokens.
        Returns:
            `~torch.Tensor`: The log likelihood. This will have size ``(batch_size,)`` if
            reduction is ``none``, ``()`` otherwise.
        """
        if reduction not in ('none', 'sum', 'mean', 'token_mean'):
            raise ValueError(f'invalid reduction: {reduction}')
        if mask is None:
            mask = torch.ones_like(tags, dtype=torch.uint8, device=tags.device)
        if mask.dtype != torch.uint8:
            mask = mask.byte()

        # calculate pRisk
        one_hot_targets = nn.functional.one_hot(tags, self.num_tags)
        unit_risk = torch.sum(torch.pow(emissions - one_hot_targets, 2), dim=-1)
        pRisk = 0.0
        for tag in self.new_tag:
            tag_id = self.label2id[tag]
            pMask = mask * (tags == tag_id)
            risk = self.masked_sum(unit_risk, pMask)
            pRisk += self.balance[tag] * risk

        # calculate nRisk: nRisk = uRisk - self.prior * (1 - pRisk)
        one_hot_targets = nn.functional.one_hot(torch.zeros_like(tags), self.num_tags)
        unit_risk = torch.sum(torch.pow(emissions - one_hot_targets, 2), dim=-1)
        _pRisk = 0.0
        for tag in self.new_tag:
            tag_id = self.label2id[tag]
            pMask = mask * (tags == tag_id)
            risk = self.masked_sum(unit_risk, pMask)
            _pRisk += self.prior[tag] * risk

        uMask = mask * (tags == 0)  # self.label2id['O'] = 0
        uRisk = self.masked_sum(unit_risk, uMask)

        nRisk = uRisk - _pRisk

        # calculate final_risk
        if nRisk < 0:
            final_risk = -nRisk
        else:
            final_risk = self.gamma * pRisk + nRisk

        return final_risk
